import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem


class SRN(Problem):
    def __init__(self):
        super().__init__(n_var=2, n_obj=2, n_ieq_constr=2, xl=-20, xu=+20, vtype=float)

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = 2 + (x[:, 0] - 2) ** 2 + (x[:, 1] - 1) ** 2
        f2 = 9 * x[:, 0] - (x[:, 1] - 1) ** 2

        g1 = x[:, 0] ** 2 + x[:, 1] ** 2 - 225
        g2 = x[:, 0] - 3 * x[:, 1] + 10

        out["F"] = anp.column_stack([f1, f2])
        out["G"] = anp.column_stack([g1, g2])

    def _calc_pareto_front(self, *args, n_points=100, **kwargs):
        ps = self.pareto_set(n_points=n_points)
        return self.evaluate(ps, return_values_of=["F"])

    def _calc_pareto_set(self, *args, n_points=100, **kwargs):
        x1 = np.full(n_points, -2.5)
        x2 = np.linspace(2.5, 14.7902, n_points)
        return np.column_stack([x1, x2])



import numpy as np

from pymoo.core.sampling import Sampling
from pymoo.util.misc import cdist


def criterion_maxmin(X):
    D = cdist(X, X)
    np.fill_diagonal(D, np.inf)
    return np.min(D)


def criterion_corr(X):
    M = np.corrcoef(X.T, rowvar=True)
    return -np.sum(np.tril(M, -1) ** 2)


def sampling_lhs(n_samples, n_var, xl=0, xu=1, smooth=True, criterion=criterion_maxmin, n_iter=50):

    X = sampling_lhs_unit(n_samples, n_var, smooth=smooth)

    # if a criterion is selected to further improve the sampling
    if criterion is not None:

        # current best score is stored here
        score = criterion(X)

        for j in range(1, n_iter):

            # create new random sample and check the score again
            _X = sampling_lhs_unit(n_samples, n_var, smooth=smooth)
            _score = criterion(_X)

            if _score > score:
                X, score = _X, _score

    return xl + X * (xu - xl)


def sampling_lhs_unit(n_samples, n_var, smooth=True):
    X = np.random.random(size=(n_samples, n_var))
    Xp = X.argsort(axis=0) + 1

    if smooth:
        Xp = Xp - np.random.random(Xp.shape)
    else:
        Xp = Xp - 0.5
    Xp /= n_samples
    return Xp


class LatinHypercubeSampling(Sampling):

    def __init__(self,
                 smooth=True,
                 iterations=20,
                 criterion=criterion_maxmin) -> None:
        super().__init__()
        self.smooth = smooth
        self.iterations = iterations
        self.criterion = criterion

    def _do(self, problem, n_samples, **kwargs):
        xl, xu = problem.bounds()

        X = sampling_lhs(n_samples, problem.n_var, xl=xl, xu=xu, smooth=self.smooth,
                         criterion=self.criterion, n_iter=self.iterations)

        return X


class LHS(LatinHypercubeSampling):
    pass



import numpy as np

from pymoo.indicators.hv.exact import DynamicHypervolume


def alpha(N, k):
    alpha = np.zeros(N+1)

    for i in range(1, N + 1):
        alpha[i] = np.prod([(k - j) / (N - j) for j in range(1, i)]) / i

    return alpha


def hv_monte_carlo(dom, V, n_dom=None):
    N, n_samples = dom.shape
    if n_dom is None:
        n_dom = dom.sum(axis=0)

    a = alpha(N, N)
    hv = sum([a[n_dom[dom[i]]].sum() for i in range(N)]) / n_samples * V
    return hv


def hvc_monte_carlo(dom, V, n_dom=None, k=1):
    N, n_samples = dom.shape
    if n_dom is None:
        n_dom = dom.sum(axis=0)

    a = alpha(N, k)
    hvc = np.array([(a[n_dom[dom[i]]].sum() / n_samples * V).sum() for i in range(N)])
    return hvc


class ApproximateMonteCarloHypervolume(DynamicHypervolume):

    def __init__(self, ref_point, n_samples=10000, n_exclusive=1, **kwargs) -> None:
        self.n_samples = n_samples
        self.n_exclusive = n_exclusive

        self.V = None
        self.dom = None

        super().__init__(ref_point, **kwargs)

    def _calc(self, ref_point, F):
        (N, M) = F.shape

        ideal = F.min(axis=0)
        V = np.prod(ref_point - ideal)

        S = np.random.uniform(low=ideal, high=ref_point, size=(self.n_samples, M))

        dom = np.array([np.all(F[i] <= S, axis=1) for i in range(N)])

        n_dom = dom.sum(axis=0)
        hv = hv_monte_carlo(dom, V, n_dom=n_dom)
        hvc = hvc_monte_carlo(dom, V, n_dom=n_dom, k=self.n_exclusive)

        self.V = V
        self.dom = dom

        return hv, hvc

    def delete(self, k):
        self.F = np.delete(self.F, k, axis=0)
        self.dom = np.delete(self.dom, k, axis=0)

        self.hv -= self.hvc[k]

        V, dom = self.V, self.dom
        n_dom = dom.sum(axis=0)
        self.hvc = hvc_monte_carlo(dom, V, n_dom=n_dom, k=self.n_exclusive)




import math

from alive_progress import alive_bar


class ProgressBar:

    def __init__(self, *args, start=True, non_decreasing=True, **kwargs):
        self.args = args
        self.kwargs = kwargs

        for key, default in [("manual", True), ("force_tty", True)]:
            if key not in kwargs:
                kwargs[key] = default

        self.func = None
        self.obj = None
        self.non_decreasing = non_decreasing
        self._max = 0.0

        if start:
            self.start()

    def set(self, value, *args, **kwargs):
        if self.non_decreasing:
            self._max = max(self._max, value)
            value = self._max

        prec = 100
        value = math.floor(value * prec) / prec

        self.obj(value, *args, **kwargs)

    def start(self):

        if not self.obj:
            # save the generator to this object
            self.func = alive_bar(*self.args, **self.kwargs).gen

            # create the bar
            self.obj = next(self.func)

    def close(self):
        if self.obj:
            try:
                next(self.func)
            except:
                pass

    def __enter__(self):
        self.start()

    def __exit__(self, type, value, traceback):
        self.close()



import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.core.population import Population
from pymoo.docs import parse_doc_string
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.termination.default import DefaultSingleObjectiveTermination
from pymoo.util.display.single import SingleObjectiveOutput
from pymoo.util.optimum import filter_optimum


class ES(GeneticAlgorithm):

    def __init__(self,
                 n_offsprings=200,
                 pop_size=None,
                 rule=1.0 / 7.0,
                 phi=1.0,
                 gamma=0.85,
                 sampling=FloatRandomSampling(),
                 survival=FitnessSurvival(),
                 output=SingleObjectiveOutput(),
                 **kwargs):

        """
        Evolutionary Strategy (ES)

        Parameters
        ----------
        n_offsprings : int
            The number of individuals created in each iteration.
        pop_size : int
            The number of individuals which are surviving from the offspring population (non-elitist)
        rule : float
            The rule (ratio) of individuals surviving. This automatically either calculated `n_offsprings` or `pop_size`.
        phi : float
            Expected rate of convergence (usually 1.0).
        gamma : float
            If not `None`, some individuals are created using the differentials with this as a length scale.
        sampling : object
            The sampling method for creating the initial population.
        """

        if pop_size is None and n_offsprings is not None:
            pop_size = int(np.math.ceil(n_offsprings * rule))
        elif n_offsprings is None and pop_size is not None:
            n_offsprings = int(np.math.floor(n_offsprings / rule))

        assert pop_size is not None and n_offsprings is not None, "You have to at least provivde pop_size of n_offsprings."
        assert n_offsprings >= 2 * pop_size, "The number of offsprings should be at least double the population size."

        super().__init__(pop_size=pop_size,
                         n_offsprings=n_offsprings,
                         sampling=sampling,
                         survival=survival,
                         output=output,
                         advance_after_initial_infill=True,
                         **kwargs)

        self.termination = DefaultSingleObjectiveTermination()
        self.phi = phi
        self.gamma = gamma

        self.tau, self.taup, self.sigma_max = None, None, None

    def _setup(self, problem, **kwargs):
        n = problem.n_var
        self.taup = self.phi / ((2 * n) ** 0.5)
        self.tau = self.phi / ((2 * (n ** 0.5)) ** 0.5)

        xl, xu = self.problem.bounds()
        self.sigma_max = (xu - xl) / (self.problem.n_var ** 0.5)

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills=infills, **kwargs)

        # initialize all individuals with the maximum sigma value
        infills.set("sigma", [self.sigma_max] * len(infills))

    def _infill(self):
        pop, mu, _lambda = self.pop, self.pop_size, self.n_offsprings
        xl, xu = self.problem.bounds()
        X, sigma = pop.get("X", "sigma")

        # cycle through the elites individuals for create the solutions
        I = np.arange(_lambda) % mu

        # transform X and sigma to the shape of number of offsprings
        X, sigma = X[I], sigma[I]

        # get the sigma only of the elites to be used
        sigmap = es_intermediate_recomb(sigma)

        # calculate the new sigma based on tau and tau prime
        sigmap = np.minimum(self.sigma_max, es_sigma(sigmap, self.tau, self.taup))

        # execute the evolutionary strategy to calculate the offspring solutions
        Xp = X + sigmap * np.random.normal(size=sigmap.shape)

        # if gamma is not none do the differential variation overwrite Xp and sigmap for the first mu-1 individuals
        if self.gamma is not None:
            Xp[:mu - 1] = X[:mu - 1] + self.gamma * (X[0] - X[1:mu])
            sigmap[:mu - 1] = sigma[:mu - 1]

        # if we have bounds to consider -> repair the individuals which are out of bounds
        if self.problem.has_bounds():
            Xp = es_mut_repair(Xp, X, sigmap, xl, xu, 10)

        # create the population to proceed further
        off = Population.new(X=Xp, sigma=sigmap)

        return off

    def _advance(self, infills=None, **kwargs):

        # if not all solutions suggested by infill() are evaluated we create a more semi (mu+lambda) algorithm
        if len(infills) < self.pop_size:
            infills = Population.merge(infills, self.pop)

        self.pop = self.survival.do(self.problem, infills, n_survive=self.pop_size)

    def _set_optimum(self):
        pop = self.pop if self.opt is None else Population.merge(self.opt, self.pop)
        self.opt = filter_optimum(pop, least_infeasible=True)


def es_sigma(sigma, tau, taup):
    _lambda, _n = sigma.shape
    return sigma * np.exp(taup * np.random.normal(size=(_lambda, 1)) + tau * np.random.normal(size=(_lambda, _n)))


def es_intermediate_recomb(sigma):
    _lambda, _n = sigma.shape
    sigma_hat = np.zeros_like(sigma)

    for i in range(_lambda):
        for j in range(_n):
            k = np.random.randint(_lambda)
            sigma_hat[i, j] = (sigma[i, j] + sigma[k, j]) / 2.0

    return sigma_hat


def es_mut_repair(Xp, X, sigma, xl, xu, n_trials):
    # reshape xl and xu to be the same shape as the input
    XL = xl[None, :].repeat(len(Xp), axis=0)
    XU = xu[None, :].repeat(len(Xp), axis=0)

    all_in_bounds = False

    # for the given number of trials
    for k in range(n_trials):

        # find all indices which are out of bounds
        i, j = np.where(np.logical_or(Xp < XL, Xp > XU))

        if len(i) == 0:
            all_in_bounds = True
            break
        else:
            # do the mutation again vectored for all values not in bound
            Xp[i, j] = X[i, j] + sigma[i, j] * np.random.normal(size=len(i))

    # if there are still solutions which boundaries are violated, set them to the original X
    if not all_in_bounds:
        i, j = np.where(np.logical_or(Xp < XL, Xp > XU))
        Xp[i, j] = X[i, j]

    return Xp


def es_mut_loop(X, sigmap, xl, xu, n_trials=10):
    _lambda, _n = sigmap.shape

    # X prime which will be returned by the algorithm (set the default value to the same as parent)
    Xp = np.zeros_like(sigmap)

    # for each of the new offsprings
    for i in range(_lambda):

        # for each variable of it
        for j in range(_n):

            # by default just copy the value if no value is in bounds this will stay
            Xp[i, j] = X[i, j]

            # try to set the value a few time and be done if in bounds
            for _ in range(n_trials):

                # calculate the mutated value
                x = X[i, j] + sigmap[i, j] * np.random.normal()

                # if it is inside the bounds accept it - otherwise try again
                if xl[j] <= x <= xu[j]:
                    Xp[i, j] = x
                    break

    return Xp


parse_doc_string(ES.__init__)



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem


class Zakharov(Problem):
    def __init__(self, n_var=2):
        super().__init__(n_var=n_var, n_obj=1, xl=-10, xu=10, vtype=float)

    def _evaluate(self, x, out, *args, **kwargs):
        a = anp.sum(0.5 * anp.arange(1, self.n_var + 1) * x, axis=1)
        out["F"] = anp.sum(anp.square(x), axis=1) + anp.square(a) + anp.power(a, 4)

    def _calc_pareto_front(self):
        return 0

    def _calc_pareto_set(self):
        return np.full(self.n_var, 0)



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem


class Truss2D(Problem):

    def __init__(self):
        super().__init__(n_var=3, n_obj=2, n_ieq_constr=1, vtype=float)

        self.Amax = 0.01
        self.Smax = 1e5

        self.xl = np.array([0.0, 0.0, 1.0])
        self.xu = np.array([self.Amax, self.Amax, 3.0])

    def _evaluate(self, x, out, *args, **kwargs):

        # variable names for convenient access
        x1 = x[:, 0]
        x2 = x[:, 1]
        y = x[:, 2]

        # first objectives
        f1 = x1 * anp.sqrt(16 + anp.square(y)) + x2 * anp.sqrt((1 + anp.square(y)))

        # measure which are needed for the second objective
        sigma_ac = 20 * anp.sqrt(16 + anp.square(y)) / (y * x1)
        sigma_bc = 80 * anp.sqrt(1 + anp.square(y)) / (y * x2)

        # take the max
        f2 = anp.max(anp.column_stack((sigma_ac, sigma_bc)), axis=1)

        # define a constraint
        g1 = f2 - self.Smax

        out["F"] = anp.column_stack([f1, f2])
        out["G"] = g1

    def _calc_pareto_front(self, *args, **kwargs):

        T = 2 * np.sqrt(5) * self.Amax

        # Part A - before transition point T

        f1 = np.linspace(400 / self.Smax, T, 1000)
        f2 = 400 / f1
        part_a = np.column_stack([f1, f2])

        # Part B - after transition point T

        def calc_y(V):
            return np.sqrt(3200 * V**2 + 40 * V * np.sqrt(6400 * V**2 - 12) - 4)

        def calc_SV(y):
            return (4 + y**2) / (0.01 * y)

        def calc_x1(y):
            return 0.0025 * np.sqrt((16 + y**2) / (1 + y**2))

        def calc_S(V):
            y = calc_y(V)
            SV = calc_SV(y)
            S = SV / V
            return S

        y = 3
        x1 = calc_x1(y)
        x2 = 0.01

        V_min = T
        V_max = self.evaluate(np.array([x1, x2, y]), return_values_of=["F"])[0]

        f1 = np.linspace(V_min, V_max, 100)
        f2 = calc_S(f1)
        part_b = np.column_stack([f1, f2])

        pf = np.row_stack([part_a, part_b])

        return pf





import numpy as np

from pymoo.core.problem import Problem
from pymoo.problems.many import generic_sphere, get_ref_dirs
from pymoo.util.function_loader import load_function
from pymoo.util.misc import powerset


class WFG(Problem):

    def __init__(self, n_var, n_obj, k=None, l=None, **kwargs):
        super().__init__(n_var=n_var,
                         n_obj=n_obj,
                         xl=0.0,
                         xu=2 * np.arange(1, n_var + 1).astype(float),
                         vtype=float,
                         **kwargs)

        self.S = np.arange(2, 2 * self.n_obj + 1, 2).astype(float)
        self.A = np.ones(self.n_obj - 1)

        if k:
            self.k = k
        else:
            if n_obj == 2:
                self.k = 4
            else:
                self.k = 2 * (n_obj - 1)

        if l:
            self.l = l
        else:
            self.l = n_var - self.k

        self.validate(self.l, self.k, self.n_obj)

    def validate(self, l, k, n_obj):
        if n_obj < 2:
            raise ValueError('WFG problems must have two or more objectives.')
        if not k % (n_obj - 1) == 0:
            raise ValueError('Position parameter (k) must be divisible by number of objectives minus one.')
        if k < 4:
            raise ValueError('Position parameter (k) must be greater or equal than 4.')
        if (k + l) < n_obj:
            raise ValueError('Sum of distance and position parameters must be greater than num. of objs. (k + l >= M).')

    def _post(self, t, a):
        x = []
        for i in range(t.shape[1] - 1):
            x.append(np.maximum(t[:, -1], a[i]) * (t[:, i] - 0.5) + 0.5)
        x.append(t[:, -1])
        return np.column_stack(x)

    def _calculate(self, x, s, h):
        return x[:, -1][:, None] + s * np.column_stack(h)

    def _rand_optimal_position(self, n):
        return np.random.random((n, self.k))

    def _positional_to_optimal(self, K):
        suffix = np.full((len(K), self.l), 0.35)
        X = np.column_stack([K, suffix])
        return X * self.xu

    def _calc_pareto_set_extremes(self):
        ps = np.ones((2 ** self.k, self.k))
        for i, s in enumerate(powerset(np.arange(self.k))):
            ps[i, s] = 0
        return self._positional_to_optimal(ps)

    def _calc_pareto_set_interior(self, n_points):
        return self._positional_to_optimal(self._rand_optimal_position(n_points))

    def _calc_pareto_set(self, n_points=500, *args, **kwargs):
        extremes = self._calc_pareto_set_extremes()
        interior = self._calc_pareto_set_interior(n_points - len(extremes))
        return np.row_stack([extremes, interior])

    def _calc_pareto_front(self, ref_dirs=None, n_iterations=200, points_each_iteration=200, *args, **kwargs):
        pf = self.evaluate(self._calc_pareto_set_extremes(), return_values_of=["F"])

        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)

        for k in range(n_iterations):
            _pf = self.evaluate(self._calc_pareto_set_interior(points_each_iteration), return_values_of=["F"])
            pf = np.row_stack([pf, _pf])

            ideal, nadir = pf.min(axis=0), pf.max(axis=0)

            N = (pf - ideal) / (nadir-ideal)
            dist_matrix = load_function("calc_perpendicular_distance")(N, ref_dirs)

            closest = np.argmin(dist_matrix, axis=0)
            pf = pf[closest]

        pf = pf[np.lexsort(pf.T[::-1])]
        return pf


class WFG1(WFG):

    @staticmethod
    def t1(x, n, k):
        x[:, k:n] = _transformation_shift_linear(x[:, k:n], 0.35)
        return x

    @staticmethod
    def t2(x, n, k):
        x[:, k:n] = _transformation_bias_flat(x[:, k:n], 0.8, 0.75, 0.85)
        return x

    @staticmethod
    def t3(x, n):
        x[:, :n] = _transformation_bias_poly(x[:, :n], 0.02)
        return x

    @staticmethod
    def t4(x, m, n, k):
        w = np.arange(2, 2 * n + 1, 2)
        gap = k // (m - 1)
        t = []
        for m in range(1, m):
            _y = x[:, (m - 1) * gap: (m * gap)]
            _w = w[(m - 1) * gap: (m * gap)]
            t.append(_reduction_weighted_sum(_y, _w))
        t.append(_reduction_weighted_sum(x[:, k:n], w[k:n]))
        return np.column_stack(t)

    def _evaluate(self, x, out, *args, **kwargs):
        y = x / self.xu
        y = WFG1.t1(y, self.n_var, self.k)
        y = WFG1.t2(y, self.n_var, self.k)
        y = WFG1.t3(y, self.n_var)
        y = WFG1.t4(y, self.n_obj, self.n_var, self.k)

        y = self._post(y, self.A)

        h = [_shape_convex(y[:, :-1], m + 1) for m in range(self.n_obj - 1)]
        h.append(_shape_mixed(y[:, 0], alpha=1.0, A=5.0))

        out["F"] = self._calculate(y, self.S, h)

    def _rand_optimal_position(self, n):
        return np.power(np.random.random((n, self.k)), 50.0)


class WFG2(WFG):

    def validate(self, l, k, n_obj):
        super().validate(l, k, n_obj)
        validate_wfg2_wfg3(l)

    @staticmethod
    def t2(x, n, k):
        y = [x[:, i] for i in range(k)]

        l = n - k
        ind_non_sep = k + l // 2

        i = k + 1
        while i <= ind_non_sep:
            head = k + 2 * (i - k) - 2
            tail = k + 2 * (i - k)
            y.append(_reduction_non_sep(x[:, head:tail], 2))
            i += 1

        return np.column_stack(y)

    @staticmethod
    def t3(x, m, n, k):
        ind_r_sum = k + (n - k) // 2
        gap = k // (m - 1)

        t = [_reduction_weighted_sum_uniform(x[:, (m - 1) * gap: (m * gap)]) for m in range(1, m)]
        t.append(_reduction_weighted_sum_uniform(x[:, k:ind_r_sum]))

        return np.column_stack(t)

    def _evaluate(self, x, out, *args, **kwargs):
        y = x / self.xu
        y = WFG1.t1(y, self.n_var, self.k)
        y = WFG2.t2(y, self.n_var, self.k)
        y = WFG2.t3(y, self.n_obj, self.n_var, self.k)
        y = self._post(y, self.A)

        h = [_shape_convex(y[:, :-1], m + 1) for m in range(self.n_obj - 1)]
        h.append(_shape_disconnected(y[:, 0], alpha=1.0, beta=1.0, A=5.0))

        out["F"] = self._calculate(y, self.S, h)


class WFG3(WFG):

    def __init__(self, n_var, n_obj, k=None, **kwargs):
        super().__init__(n_var, n_obj, k=k, **kwargs)
        self.A[1:] = 0

    def validate(self, l, k, n_obj):
        super().validate(l, k, n_obj)
        validate_wfg2_wfg3(l)

    def _evaluate(self, x, out, *args, **kwargs):
        y = x / self.xu
        y = WFG1.t1(y, self.n_var, self.k)
        y = WFG2.t2(y, self.n_var, self.k)
        y = WFG2.t3(y, self.n_obj, self.n_var, self.k)
        y = self._post(y, self.A)

        h = [_shape_linear(y[:, :-1], m + 1) for m in range(self.n_obj)]

        out["F"] = self._calculate(y, self.S, h)

    # def _calc_pareto_front(self, ref_dirs=None):
    #     if ref_dirs is None:
    #         ref_dirs = get_ref_dirs(self.n_obj)
    #     return ref_dirs * self.S


class WFG4(WFG):

    @staticmethod
    def t1(x):
        return _transformation_shift_multi_modal(x, 30.0, 10.0, 0.35)

    @staticmethod
    def t2(x, m, k):
        gap = k // (m - 1)
        t = [_reduction_weighted_sum_uniform(x[:, (m - 1) * gap: (m * gap)]) for m in range(1, m)]
        t.append(_reduction_weighted_sum_uniform(x[:, k:]))
        return np.column_stack(t)

    def _evaluate(self, x, out, *args, **kwargs):
        y = x / self.xu
        y = WFG4.t1(y)
        y = WFG4.t2(y, self.n_obj, self.k)
        y = self._post(y, self.A)

        h = [_shape_concave(y[:, :-1], m + 1) for m in range(self.n_obj)]

        out["F"] = self._calculate(y, self.S, h)

    # def _calc_pareto_front(self, ref_dirs=None):
    #     if ref_dirs is None:
    #         ref_dirs = get_ref_dirs(self.n_obj)
    #     return generic_sphere(ref_dirs) * self.S


class WFG5(WFG):

    @staticmethod
    def t1(x):
        return _transformation_param_deceptive(x, A=0.35, B=0.001, C=0.05)

    def _evaluate(self, x, out, *args, **kwargs):
        y = x / self.xu
        y = WFG5.t1(y)
        y = WFG4.t2(y, self.n_obj, self.k)
        y = self._post(y, self.A)

        h = [_shape_concave(y[:, :-1], m + 1) for m in range(self.n_obj)]

        out["F"] = self._calculate(y, self.S, h)

    # def _calc_pareto_front(self, ref_dirs=None):
    #     if ref_dirs is None:
    #         ref_dirs = get_ref_dirs(self.n_obj)
    #     return generic_sphere(ref_dirs) * self.S


class WFG6(WFG):

    @staticmethod
    def t2(x, m, n, k):
        gap = k // (m - 1)
        t = [_reduction_non_sep(x[:, (m - 1) * gap: (m * gap)], gap) for m in range(1, m)]
        t.append(_reduction_non_sep(x[:, k:], n - k))
        return np.column_stack(t)

    def _evaluate(self, x, out, *args, **kwargs):
        y = x / self.xu
        y = WFG1.t1(y, self.n_var, self.k)
        y = WFG6.t2(y, self.n_obj, self.n_var, self.k)
        y = self._post(y, self.A)

        h = [_shape_concave(y[:, :-1], m + 1) for m in range(self.n_obj)]

        out["F"] = self._calculate(y, self.S, h)

    # def _calc_pareto_front(self, ref_dirs=None):
    #     if ref_dirs is None:
    #         ref_dirs = get_ref_dirs(self.n_obj)
    #     return generic_sphere(ref_dirs) * self.S


class WFG7(WFG):

    @staticmethod
    def t1(x, k):
        for i in range(k):
            aux = _reduction_weighted_sum_uniform(x[:, i + 1:])
            x[:, i] = _transformation_param_dependent(x[:, i], aux)
        return x

    def _evaluate(self, x, out, *args, **kwargs):
        y = x / self.xu
        y = WFG7.t1(y, self.k)
        y = WFG1.t1(y, self.n_var, self.k)
        y = WFG4.t2(y, self.n_obj, self.k)
        y = self._post(y, self.A)

        h = [_shape_concave(y[:, :-1], m + 1) for m in range(self.n_obj)]

        out["F"] = self._calculate(y, self.S, h)

    # def _calc_pareto_front(self, ref_dirs=None):
    #     if ref_dirs is None:
    #         ref_dirs = get_ref_dirs(self.n_obj)
    #     return generic_sphere(ref_dirs) * self.S


class WFG8(WFG):

    @staticmethod
    def t1(x, n, k):
        ret = []
        for i in range(k, n):
            aux = _reduction_weighted_sum_uniform(x[:, :i])
            ret.append(_transformation_param_dependent(x[:, i], aux, A=0.98 / 49.98, B=0.02, C=50.0))
        return np.column_stack(ret)

    def _evaluate(self, x, out, *args, **kwargs):
        y = x / self.xu
        y[:, self.k:self.n_var] = WFG8.t1(y, self.n_var, self.k)
        y = WFG1.t1(y, self.n_var, self.k)
        y = WFG4.t2(y, self.n_obj, self.k)
        y = self._post(y, self.A)

        h = [_shape_concave(y[:, :-1], m + 1) for m in range(self.n_obj)]

        out["F"] = self._calculate(y, self.S, h)

    def _positional_to_optimal(self, K):
        k, l = self.k, self.l

        for i in range(k, k + l):
            u = K.sum(axis=1) / K.shape[1]
            tmp1 = np.abs(np.floor(0.5 - u) + 0.98 / 49.98)
            tmp2 = 0.02 + 49.98 * (0.98 / 49.98 - (1.0 - 2.0 * u) * tmp1)
            suffix = np.power(0.35, np.power(tmp2, -1.0))

            K = np.column_stack([K, suffix[:, None]])

        ret = K * (2 * (np.arange(self.n_var) + 1))
        return ret


class WFG9(WFG):

    @staticmethod
    def t1(x, n):
        ret = []
        for i in range(0, n - 1):
            aux = _reduction_weighted_sum_uniform(x[:, i + 1:])
            ret.append(_transformation_param_dependent(x[:, i], aux))
        return np.column_stack(ret)

    @staticmethod
    def t2(x, n, k):
        a = [_transformation_shift_deceptive(x[:, i], 0.35, 0.001, 0.05) for i in range(k)]
        b = [_transformation_shift_multi_modal(x[:, i], 30.0, 95.0, 0.35) for i in range(k, n)]
        return np.column_stack(a + b)

    @staticmethod
    def t3(x, m, n, k):
        gap = k // (m - 1)
        t = [_reduction_non_sep(x[:, (m - 1) * gap: (m * gap)], gap) for m in range(1, m)]
        t.append(_reduction_non_sep(x[:, k:], n - k))
        return np.column_stack(t)

    def _evaluate(self, x, out, *args, **kwargs):
        y = x / self.xu
        y[:, :self.n_var - 1] = WFG9.t1(y, self.n_var)
        y = WFG9.t2(y, self.n_var, self.k)
        y = WFG9.t3(y, self.n_obj, self.n_var, self.k)

        h = [_shape_concave(y[:, :-1], m + 1) for m in range(self.n_obj)]

        out["F"] = self._calculate(y, self.S, h)

    def _positional_to_optimal(self, K):
        k, l = self.k, self.l

        suffix = np.full((len(K), self.l), 0.0)
        X = np.column_stack([K, suffix])
        X[:, self.k + self.l - 1] = 0.35

        for i in range(self.k + self.l - 2, self.k - 1, -1):
            m = X[:, i + 1:k + l]
            val = m.sum(axis=1) / m.shape[1]
            X[:, i] = 0.35 ** ((0.02 + 1.96 * val) ** -1)

        ret = X * (2 * (np.arange(self.n_var) + 1))
        return ret

    def _calc_pareto_front(self, ref_dirs=None):
        if ref_dirs is None:
            ref_dirs = get_ref_dirs(self.n_obj)
        return generic_sphere(ref_dirs) * self.S


# ---------------------------------------------------------------------------------------------------------
# TRANSFORMATIONS
# ---------------------------------------------------------------------------------------------------------


def _transformation_shift_linear(value, shift=0.35):
    return correct_to_01(np.fabs(value - shift) / np.fabs(np.floor(shift - value) + shift))


def _transformation_shift_deceptive(y, A=0.35, B=0.005, C=0.05):
    tmp1 = np.floor(y - A + B) * (1.0 - C + (A - B) / B) / (A - B)
    tmp2 = np.floor(A + B - y) * (1.0 - C + (1.0 - A - B) / B) / (1.0 - A - B)
    ret = 1.0 + (np.fabs(y - A) - B) * (tmp1 + tmp2 + 1.0 / B)
    return correct_to_01(ret)


def _transformation_shift_multi_modal(y, A, B, C):
    tmp1 = np.fabs(y - C) / (2.0 * (np.floor(C - y) + C))
    tmp2 = (4.0 * A + 2.0) * np.pi * (0.5 - tmp1)
    ret = (1.0 + np.cos(tmp2) + 4.0 * B * np.power(tmp1, 2.0)) / (B + 2.0)
    return correct_to_01(ret)


def _transformation_bias_flat(y, a, b, c):
    ret = a + np.minimum(0, np.floor(y - b)) * (a * (b - y) / b) \
          - np.minimum(0, np.floor(c - y)) * ((1.0 - a) * (y - c) / (1.0 - c))
    return correct_to_01(ret)


def _transformation_bias_poly(y, alpha):
    return correct_to_01(y ** alpha)


def _transformation_param_dependent(y, y_deg, A=0.98 / 49.98, B=0.02, C=50.0):
    aux = A - (1.0 - 2.0 * y_deg) * np.fabs(np.floor(0.5 - y_deg) + A)
    ret = np.power(y, B + (C - B) * aux)
    return correct_to_01(ret)


def _transformation_param_deceptive(y, A=0.35, B=0.001, C=0.05):
    tmp1 = np.floor(y - A + B) * (1.0 - C + (A - B) / B) / (A - B)
    tmp2 = np.floor(A + B - y) * (1.0 - C + (1.0 - A - B) / B) / (1.0 - A - B)
    ret = 1.0 + (np.fabs(y - A) - B) * (tmp1 + tmp2 + 1.0 / B)
    return correct_to_01(ret)


# ---------------------------------------------------------------------------------------------------------
# REDUCTION
# ---------------------------------------------------------------------------------------------------------


def _reduction_weighted_sum(y, w):
    return correct_to_01(np.dot(y, w) / w.sum())


def _reduction_weighted_sum_uniform(y):
    return correct_to_01(y.mean(axis=1))


def _reduction_non_sep(y, A):
    n, m = y.shape
    val = np.ceil(A / 2.0)

    num = np.zeros(n)
    for j in range(m):
        num += y[:, j]
        for k in range(A - 1):
            num += np.fabs(y[:, j] - y[:, (1 + j + k) % m])

    denom = m * val * (1.0 + 2.0 * A - 2 * val) / A

    return correct_to_01(num / denom)


# ---------------------------------------------------------------------------------------------------------
# SHAPE
# ---------------------------------------------------------------------------------------------------------


def _shape_concave(x, m):
    M = x.shape[1]
    if m == 1:
        ret = np.prod(np.sin(0.5 * x[:, :M] * np.pi), axis=1)
    elif 1 < m <= M:
        ret = np.prod(np.sin(0.5 * x[:, :M - m + 1] * np.pi), axis=1)
        ret *= np.cos(0.5 * x[:, M - m + 1] * np.pi)
    else:
        ret = np.cos(0.5 * x[:, 0] * np.pi)
    return correct_to_01(ret)


def _shape_convex(x, m):
    M = x.shape[1]
    if m == 1:
        ret = np.prod(1.0 - np.cos(0.5 * x[:, :M] * np.pi), axis=1)
    elif 1 < m <= M:
        ret = np.prod(1.0 - np.cos(0.5 * x[:, :M - m + 1] * np.pi), axis=1)
        ret *= 1.0 - np.sin(0.5 * x[:, M - m + 1] * np.pi)
    else:
        ret = 1.0 - np.sin(0.5 * x[:, 0] * np.pi)
    return correct_to_01(ret)


def _shape_linear(x, m):
    M = x.shape[1]
    if m == 1:
        ret = np.prod(x, axis=1)
    elif 1 < m <= M:
        ret = np.prod(x[:, :M - m + 1], axis=1)
        ret *= 1.0 - x[:, M - m + 1]
    else:
        ret = 1.0 - x[:, 0]
    return correct_to_01(ret)


def _shape_mixed(x, A=5.0, alpha=1.0):
    aux = 2.0 * A * np.pi
    ret = np.power(1.0 - x - (np.cos(aux * x + 0.5 * np.pi) / aux), alpha)
    return correct_to_01(ret)


def _shape_disconnected(x, alpha=1.0, beta=1.0, A=5.0):
    aux = np.cos(A * np.pi * x ** beta)
    return correct_to_01(1.0 - x ** alpha * aux ** 2)


# ---------------------------------------------------------------------------------------------------------
# UTIL
# ---------------------------------------------------------------------------------------------------------

def validate_wfg2_wfg3(l):
    if not l % 2 == 0:
        raise ValueError('In WFG2/WFG3 the distance-related parameter (l) must be divisible by 2.')


def correct_to_01(X, epsilon=1.0e-10):
    X[np.logical_and(X < 0, X >= 0 - epsilon)] = 0
    X[np.logical_and(X > 1, X <= 1 + epsilon)] = 1
    return X



try:
    import numba
    from numba import jit
except:
    raise Exception("Please install numba to use AGEMOEA: pip install numba")

import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.algorithms.moo.nsga2 import binary_tournament
from pymoo.core.survival import Survival
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.tournament import TournamentSelection
from pymoo.termination.default import DefaultMultiObjectiveTermination
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.misc import has_feasible
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting


# =========================================================================================================
# Implementation
# =========================================================================================================

class AGEMOEA(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=FloatRandomSampling(),
                 selection=TournamentSelection(func_comp=binary_tournament),
                 crossover=SBX(eta=15, prob=0.9),
                 mutation=PM(eta=20),
                 eliminate_duplicates=True,
                 n_offsprings=None,
                 output=MultiObjectiveOutput(),
                 **kwargs):
        """
        Adapted from:
        Panichella, A. (2019). An adaptive evolutionary algorithm based on non-euclidean geometry for many-objective
        optimization. GECCO 2019 - Proceedings of the 2019 Genetic and Evolutionary Computation Conference, July, 595–603.
        https://doi.org/10.1145/3321707.3321839

        Parameters
        ----------
        pop_size : {pop_size}
        sampling : {sampling}
        selection : {selection}
        crossover : {crossover}
        mutation : {mutation}
        eliminate_duplicates : {eliminate_duplicates}
        n_offsprings : {n_offsprings}
        """

        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         survival=AGEMOEASurvival(),
                         eliminate_duplicates=eliminate_duplicates,
                         n_offsprings=n_offsprings,
                         output=output,
                         advance_after_initial_infill=True,
                         **kwargs)
        self.default_termination = DefaultMultiObjectiveTermination()

        self.tournament_type = 'comp_by_rank_and_crowding'

    def _set_optimum(self, **kwargs):
        if not has_feasible(self.pop):
            self.opt = self.pop[[np.argmin(self.pop.get("CV"))]]
        else:
            self.opt = self.pop[self.pop.get("rank") == 0]


# ---------------------------------------------------------------------------------------------------------
# Survival Selection
# ---------------------------------------------------------------------------------------------------------

class AGEMOEASurvival(Survival):

    def __init__(self) -> None:
        super().__init__(filter_infeasible=True)
        self.nds = NonDominatedSorting()

    def _do(self, problem, pop, *args, n_survive=None, **kwargs):

        # get the objective values
        F = pop.get("F")

        N = n_survive

        # Non-dominated sorting
        fronts = self.nds.do(F, n_stop_if_ranked=N)

        # get max int value
        max_val = np.iinfo(np.int).max

        # initialize population ranks with max int value
        front_no = np.full(F.shape[0], max_val, dtype=np.int)

        # assign the rank to each individual
        for i, fr in enumerate(fronts):
            front_no[fr] = i

        pop.set("rank", front_no)

        # get the index of the front to be sorted and cut
        max_f_no = np.max(front_no[front_no != max_val])

        # keep fronts that have lower rank than the front to cut
        selected: np.ndarray = front_no < max_f_no

        n_ind, _ = F.shape

        # crowding distance is positive and has to be maximized
        crowd_dist = np.zeros(n_ind)

        # get the first front for normalization
        front1 = F[front_no == 0, :]

        # follows from the definition of the ideal point but with current non dominated solutions
        ideal_point = np.min(front1, axis=0)

        # Calculate the crowding distance of the first front as well as p and the normalization constants
        crowd_dist[front_no == 0], p, normalization = self.survival_score(front1, ideal_point)
        for i in range(1, max_f_no):  # skip first front since it is normalized by survival_score
            front = F[front_no == i, :]
            m, _ = front.shape
            front = front / normalization
            crowd_dist[front_no == i] = 1. / self.minkowski_distances(front, ideal_point[None, :], p=p).squeeze()

        # Select the solutions in the last front based on their crowding distances
        last = np.arange(selected.shape[0])[front_no == max_f_no]
        rank = np.argsort(crowd_dist[last])[::-1]
        selected[last[rank[: N - np.sum(selected)]]] = True

        pop.set("crowding", crowd_dist)

        # return selected solutions, number of selected should be equal to population size
        return pop[selected]

    def survival_score(self, front, ideal_point):
        front = np.round(front, 12, out=front)
        m, n = front.shape
        crowd_dist = np.zeros(m)

        if m < n:
            p = 1
            normalization = np.max(front, axis=0)
            return crowd_dist, p, normalization

        # shift the ideal point to the origin
        front = front - ideal_point

        # Detect the extreme points and normalize the front
        extreme = find_corner_solutions(front)
        front, normalization = normalize(front, extreme)

        # set the distance for the extreme solutions
        crowd_dist[extreme] = np.inf
        selected = np.full(m, False)
        selected[extreme] = True

        p = self.compute_geometry(front, extreme, n)

        nn = np.linalg.norm(front, p, axis=1)
        distances = self.pairwise_distances(front, p) / nn[:, None]

        neighbors = 2
        remaining = np.arange(m)
        remaining = list(remaining[~selected])
        for i in range(m - np.sum(selected)):
            mg = np.meshgrid(np.arange(selected.shape[0])[selected], remaining, copy=False, sparse=False)
            D_mg = distances[tuple(mg)]  # avoid Numpy's future deprecation of array special indexing

            if D_mg.shape[1] > 1:
                # equivalent to mink(distances(remaining, selected),neighbors,2); in Matlab
                maxim = np.argpartition(D_mg, neighbors - 1, axis=1)[:, :neighbors]
                tmp = np.sum(np.take_along_axis(D_mg, maxim, axis=1), axis=1)
                index: int = np.argmax(tmp)
                d = tmp[index]
            else:
                index = D_mg[:, 0].argmax()
                d = D_mg[index, 0]

            best = remaining.pop(index)
            selected[best] = True
            crowd_dist[best] = d

        return crowd_dist, p, normalization

    @staticmethod
    def compute_geometry(front, extreme, n):
        # approximate p(norm)
        d = point_2_line_distance(front, np.zeros(n), np.ones(n))
        d[extreme] = np.inf
        index = np.argmin(d)

        p = np.log(n) / np.log(1.0 / np.mean(front[index, :]))

        if np.isnan(p) or p <= 0.1:
            p = 1.0
        elif p > 20:
            p = 20.0  # avoid numpy underflow

        return p

    @staticmethod
    @jit(fastmath=True)
    def pairwise_distances(front, p):
        m = np.shape(front)[0]
        distances = np.zeros((m, m))
        for i in range(m):
            distances[i] = np.sum(np.abs(front[i] - front) ** p, 1) ** (1 / p)

        return distances

    @staticmethod
    @jit(fastmath=True)
    def minkowski_distances(A, B, p):
        m1 = np.shape(A)[0]
        m2 = np.shape(B)[0]
        distances = np.zeros((m1, m2))
        for i in range(m1):
            for j in range(m2):
                distances[i][j] = sum(np.abs(A[i] - B[j]) ** p) ** (1 / p)

        return distances


@jit(nopython=True, fastmath=True)
def find_corner_solutions(front):
    """Return the indexes of the extreme points"""

    m, n = front.shape

    if m <= n:
        return np.arange(m)

    # let's define the axes of the n-dimensional spaces
    W = 1e-6 + np.eye(n)
    r = W.shape[0]
    indexes = np.zeros(n, dtype=numba.intp)
    selected = np.zeros(m, dtype=numba.boolean)
    for i in range(r):
        dists = point_2_line_distance(front, np.zeros(n), W[i, :])
        dists[selected] = np.inf  # prevent already selected to be reselected
        index = np.argmin(dists)
        indexes[i] = index
        selected[index] = True
    return indexes


@jit(fastmath=True)
def point_2_line_distance(P, A, B):
    d = np.zeros(P.shape[0])

    for i in range(P.shape[0]):
        pa = P[i] - A
        ba = B - A
        t = np.dot(pa, ba) / np.dot(ba, ba)
        d[i] = sum((pa - t * ba) ** 2)

    return d


# =========================================================================================================
# Normalization
# =========================================================================================================

def normalize(front, extreme):
    m, n = front.shape

    if len(extreme) != len(np.unique(extreme, axis=0)):
        normalization = np.max(front, axis=0)
        front = front / normalization
        return front, normalization

    # Calculate the intercepts of the hyperplane constructed by the extreme
    # points and the axes

    try:
        hyperplane = np.linalg.solve(front[extreme], np.ones(n))
        if any(np.isnan(hyperplane)) or any(np.isinf(hyperplane)) or any(hyperplane < 0):
            normalization = np.max(front, axis=0)
        else:
            normalization = 1. / hyperplane
            if any(np.isnan(normalization)) or any(np.isinf(normalization)):
                normalization = np.max(front, axis=0)
    except np.linalg.LinAlgError:
        normalization = np.max(front, axis=0)

    normalization[normalization == 0.0] = 1.0

    # Normalization
    front = front / normalization

    return front, normalization

parse_doc_string(AGEMOEA.__init__)


import numpy as np

from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.core.problem import Problem


class Evaluator:

    def __init__(self,
                 skip_already_evaluated: bool = True,
                 evaluate_values_of: list = ["F", "G", "H"],
                 callback=None):

        """
        The evaluator has the purpose to glue the problem with the population/individual objects.
        Additionally, it serves as a bookkeeper to store determine the number of function evaluations of runs, time,
        and others.


        Parameters
        ----------
        skip_already_evaluated : bool
            If individual that are already evaluated shall be skipped.

        evaluate_values_of : list
            The type of values to be asked the problem to evaluated. By default all objective, ieq. and eq. constraints.

        """

        self.evaluate_values_of = evaluate_values_of
        self.skip_already_evaluated = skip_already_evaluated
        self.callback = callback

        # current number of function evaluations - initialized to zero
        self.n_eval = 0

    def eval(self,
             problem: Problem,
             pop: Population,
             skip_already_evaluated: bool = None,
             evaluate_values_of: list = None,
             count_evals: bool = True,
             **kwargs):

        # load the default settings from the evaluator object if not already provided
        evaluate_values_of = self.evaluate_values_of if evaluate_values_of is None else evaluate_values_of
        skip_already_evaluated = self.skip_already_evaluated if skip_already_evaluated is None else skip_already_evaluated

        # check the type of the input
        is_individual = isinstance(pop, Individual)

        # make sure the object is a population
        if is_individual:
            pop = Population().create(pop)

        # filter the index to have individual where not all attributes have been evaluated
        if skip_already_evaluated:
            I = [i for i, ind in enumerate(pop) if not all([e in ind.evaluated for e in evaluate_values_of])]

        # if skipping is deactivated simply make the index being all individuals
        else:
            I = np.arange(len(pop))

        # evaluate the solutions (if there are any)
        if len(I) > 0:

            # do the actual evaluation - call the sub-function to set the corresponding values to the population
            self._eval(problem, pop[I], evaluate_values_of, **kwargs)

        # update the function evaluation counter
        if count_evals:
            self.n_eval += len(I)

        # allow to have a callback registered
        if self.callback:
            self.callback(pop)

        if is_individual:
            return pop[0]
        else:
            return pop

    def _eval(self, problem, pop, evaluate_values_of, **kwargs):

        # get the design space value from the individuals
        X = pop.get("X")

        # call the problem to evaluate the solutions
        out = problem.evaluate(X, return_values_of=evaluate_values_of, return_as_dictionary=True, **kwargs)

        # for each of the attributes set it to the problem
        for key, val in out.items():
            if val is not None:
                pop.set(key, val)

        # finally set all the attributes to be evaluated for all individuals
        pop.apply(lambda ind: ind.evaluated.update(out.keys()))


class VoidEvaluator(Evaluator):

    def __init__(self, value=np.inf, **kwargs):
        super().__init__(**kwargs)
        self.value = value

    def eval(self, problem, pop, **kwargs):
        val = self.value
        if val is not None:
            for individual in pop:
                if len(individual.evaluated) == 0:
                    individual.F = np.full(problem.n_obj, val)
                    individual.G = np.full(problem.n_ieq_constr, val) if problem.n_ieq_constr > 0 else None
                    individual.H = np.full(problem.n_eq_constr, val) if problem.n_eq_constr else None
                    individual.CV = [-np.inf]
                    individual.feas = [False]



import numpy as np

from pymoo.algorithms.moo.nsga3 import calc_niche_count, niching, comp_by_cv_then_random, associate_to_niches, NSGA3
from pymoo.core.survival import Survival
from pymoo.docs import parse_doc_string
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.tournament import TournamentSelection
from pymoo.util.misc import intersect
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.util.normalization import denormalize, get_extreme_points_c, get_nadir_point
from pymoo.util.reference_direction import UniformReferenceDirectionFactory


# =========================================================================================================
# Implementation
# =========================================================================================================


class RNSGA3(NSGA3):

    def __init__(self,
                 ref_points,
                 pop_per_ref_point,
                 mu=0.05,
                 sampling=FloatRandomSampling(),
                 selection=TournamentSelection(func_comp=comp_by_cv_then_random),
                 eliminate_duplicates=True,
                 n_offsprings=None,
                 **kwargs):
        """

        Parameters
        ----------

        ref_points : {ref_points}
        
        pop_per_ref_point : int
            Size of the population used for each reference point.

        mu : float
            Defines the init_simplex_scale of the reference lines used during survival selection. Increasing mu will result
            having solutions with a larger spread.

        n_offsprings : {n_offsprings}
        sampling : {sampling}
        selection : {selection}
        crossover : {crossover}
        mutation : {mutation}
        eliminate_duplicates : {eliminate_duplicates}

        """

        # number of objectives the reference lines have
        n_obj = ref_points.shape[1]

        # add the aspiration point lines
        aspiration_ref_dirs = UniformReferenceDirectionFactory(n_dim=n_obj, n_points=pop_per_ref_point).do()

        survival = AspirationPointSurvival(ref_points, aspiration_ref_dirs, mu=mu)
        pop_size = ref_points.shape[0] * aspiration_ref_dirs.shape[0] + aspiration_ref_dirs.shape[1]
        ref_dirs = None

        super().__init__(ref_dirs,
                         pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         survival=survival,
                         eliminate_duplicates=eliminate_duplicates,
                         n_offsprings=n_offsprings,
                         **kwargs)

    def _setup(self, problem, **kwargs):
        if self.survival.ref_points.shape[1] != problem.n_obj:
            raise Exception("Dimensionality of reference points must be equal to the number of objectives: %s != %s" %
                            (self.survival.ref_points.shape[1], problem.n_obj))

    def _finalize(self):
        pass


class AspirationPointSurvival(Survival):

    def __init__(self, ref_points, aspiration_ref_dirs, mu=0.1):
        super().__init__()

        self.ref_points = ref_points
        self.aspiration_ref_dirs = aspiration_ref_dirs
        self.mu = mu

        self.ref_dirs = aspiration_ref_dirs
        self.extreme_points = None
        self.intercepts = None
        self.nadir_point = None
        self.opt = None
        self.ideal_point = np.full(ref_points.shape[1], np.inf)
        self.worst_point = np.full(ref_points.shape[1], -np.inf)

    def _do(self, problem, pop, n_survive, D=None, **kwargs):

        # attributes to be set after the survival
        F = pop.get("F")

        # find or usually update the new ideal point - from feasible solutions
        self.ideal_point = np.min(np.vstack((self.ideal_point, F, self.ref_points)), axis=0)
        self.worst_point = np.max(np.vstack((self.worst_point, F, self.ref_points)), axis=0)

        # calculate the fronts of the population
        fronts, rank = NonDominatedSorting().do(F, return_rank=True, n_stop_if_ranked=n_survive)
        non_dominated, last_front = fronts[0], fronts[-1]

        # find the extreme points for normalization
        self.extreme_points = get_extreme_points_c(
            np.vstack([F[non_dominated], self.ref_points])
            , self.ideal_point,
            extreme_points=self.extreme_points)

        # find the intercepts for normalization and do backup if gaussian elimination fails
        worst_of_population = np.max(F, axis=0)
        worst_of_front = np.max(F[non_dominated, :], axis=0)

        self.nadir_point = get_nadir_point(self.extreme_points, self.ideal_point, self.worst_point,
                                           worst_of_population, worst_of_front)

        #  consider only the population until we come to the splitting front
        I = np.concatenate(fronts)
        pop, rank, F = pop[I], rank[I], F[I]

        # update the front indices for the current population
        counter = 0
        for i in range(len(fronts)):
            for j in range(len(fronts[i])):
                fronts[i][j] = counter
                counter += 1
        last_front = fronts[-1]

        unit_ref_points = (self.ref_points - self.ideal_point) / (self.nadir_point - self.ideal_point)
        ref_dirs = get_ref_dirs_from_points(unit_ref_points, self.aspiration_ref_dirs, mu=self.mu)
        self.ref_dirs = denormalize(ref_dirs, self.ideal_point, self.nadir_point)

        # associate individuals to niches
        niche_of_individuals, dist_to_niche, dist_matrix = associate_to_niches(F, ref_dirs, self.ideal_point,
                                                                               self.nadir_point)
        pop.set('rank', rank, 'niche', niche_of_individuals, 'dist_to_niche', dist_to_niche)

        # set the optimum, first front and closest to all reference directions
        closest = np.unique(dist_matrix[:, np.unique(niche_of_individuals)].argmin(axis=0))
        self.opt = pop[intersect(fronts[0], closest)]

        # if we need to select individuals to survive
        if len(pop) > n_survive:

            # if there is only one front
            if len(fronts) == 1:
                n_remaining = n_survive
                until_last_front = np.array([], dtype=int)
                niche_count = np.zeros(len(ref_dirs), dtype=int)

            # if some individuals already survived
            else:
                until_last_front = np.concatenate(fronts[:-1])
                niche_count = calc_niche_count(len(ref_dirs), niche_of_individuals[until_last_front])
                n_remaining = n_survive - len(until_last_front)

            S = niching(pop[last_front], n_remaining, niche_count, niche_of_individuals[last_front],
                        dist_to_niche[last_front])

            survivors = np.concatenate((until_last_front, last_front[S].tolist()))
            pop = pop[survivors]

        return pop


def get_ref_dirs_from_points(ref_point, ref_dirs, mu=0.1):
    """
    This function takes user specified reference points, and creates smaller sets of equidistant
    Das-Dennis points around the projection of user points on the Das-Dennis hyperplane
    :param ref_point: List of user specified reference points
    :param n_obj: Number of objectives to consider
    :param mu: Shrinkage factor (0-1), Smaller = tigher convergence, Larger= larger convergence
    :return: Set of reference points
    """

    n_obj = ref_point.shape[1]

    val = []
    n_vector = np.ones(n_obj) / np.sqrt(n_obj)  # Normal vector of Das Dennis plane
    point_on_plane = np.eye(n_obj)[0]  # Point on Das-Dennis

    for point in ref_point:

        ref_dir_for_aspiration_point = np.copy(ref_dirs)  # Copy of computed reference directions
        ref_dir_for_aspiration_point = mu * ref_dir_for_aspiration_point

        cent = np.mean(ref_dir_for_aspiration_point, axis=0)  # Find centroid of shrunken reference points

        # Project shrunken Das-Dennis points back onto original Das-Dennis hyperplane
        intercept = line_plane_intersection(np.zeros(n_obj), point, point_on_plane, n_vector)
        shift = intercept - cent  # shift vector

        ref_dir_for_aspiration_point += shift

        # If reference directions are located outside of first octant, redefine points onto the border
        if not (ref_dir_for_aspiration_point > 0).min():
            ref_dir_for_aspiration_point[ref_dir_for_aspiration_point < 0] = 0
            ref_dir_for_aspiration_point = ref_dir_for_aspiration_point / np.sum(ref_dir_for_aspiration_point, axis=1)[
                                                                          :, None]
        val.extend(ref_dir_for_aspiration_point)

    val.extend(np.eye(n_obj))  # Add extreme points
    return np.array(val)


# intersection function

def line_plane_intersection(l0, l1, p0, p_no, epsilon=1e-6):
    """
    l0, l1: define the line
    p0, p_no: define the plane:
        p0 is a point on the plane (plane coordinate).
        p_no is a normal vector defining the plane direction;
             (does not need to be normalized).

    reference: https://en.wikipedia.org/wiki/Line%E2%80%93plane_intersection
    return a Vector or None (when the intersection can't be found).
    """

    l = l1 - l0
    dot = np.dot(l, p_no)

    if abs(dot) > epsilon:
        # the factor of the point between p0 -> p1 (0 - 1)
        # if 'fac' is between (0 - 1) the point intersects with the segment.
        # otherwise:
        #  < 0.0: behind p0.
        #  > 1.0: infront of p1.
        w = p0 - l0
        d = np.dot(w, p_no) / dot
        l = l * d
        return l0 + l
    else:
        # The segment is parallel to plane then return the perpendicular projection
        ref_proj = l1 - (np.dot(l1 - p0, p_no) * p_no)
        return ref_proj


parse_doc_string(RNSGA3.__init__)



import numpy as np

from pymoo.core.duplicate import DefaultDuplicateElimination
from pymoo.core.population import Population, merge
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting


class Truncation:

    def __call__(self, sols, k):
        pass


class RandomTruncation(Truncation):

    def __call__(self, sols, k):
        return np.random.choice(sols, size=k, replace=False)


class SurvivalTruncation(Truncation):

    def __init__(self, survival, problem=None) -> None:
        super().__init__()
        self.survival = survival

        if problem is None:
            from pymoo.core.problem import Problem
            problem = Problem()

        self.problem = problem

    def __call__(self, sols, k):
        return self.survival.do(self.problem, sols, n_survive=k)


class Archive(Population):

    def __new__(cls,
                individuals=[],
                max_size=None,
                truncate_size=None,
                truncation=RandomTruncation(),
                duplicate_elimination=DefaultDuplicateElimination(epsilon=1e-32)):

        obj = super().__new__(cls, individuals)
        obj.max_size = max_size
        obj.truncate_size = min(max_size, truncate_size) if truncate_size is not None else max_size
        obj.truncation = truncation
        obj.duplicate_elimination = duplicate_elimination

        return obj

    def __array_finalize__(self, obj):
        if obj is None:  # __new__ handles instantiation
            return

        max_size = getattr(obj, 'max_size', None)
        truncate_size = getattr(obj, 'truncate_size', None)
        truncation = getattr(obj, 'truncation', RandomTruncation())
        duplicate_elimination = getattr(obj, 'duplicate_elimination', DefaultDuplicateElimination(epsilon=1e-32))

        self.max_size = max_size
        self.truncate_size = min(max_size, truncate_size) if truncate_size is not None else max_size
        self.truncation = truncation
        self.duplicate_elimination = duplicate_elimination

    def add(self, sols):

        if len(self) > 0:
            sols = merge(self, sols)

        opt = self._find_opt(sols)

        if self.duplicate_elimination:
            opt = self.duplicate_elimination.do(opt)

        if self.max_size and len(opt) > self.max_size:
            opt = self.truncation(opt, self.truncate_size)

        cls = self.__class__
        obj = cls.__new__(cls, individuals=opt, **self.view(Archive).__dict__)
        return obj

    def _find_opt(self, sols):
        pass


class VoidArchive(Archive):

    def add(self, sols):
        return self


class SingleObjectiveArchive(Archive):

    def __new__(cls, max_size=10, **kwargs):
        return super().__new__(cls, max_size=max_size, **kwargs).view(cls)

    def _find_opt(self, sols):
        feas = sols.get("feas")

        if np.any(feas):
            sols = sols[feas]

            f = sols.get("f")
            I, = np.where(f == f[f.argmin()])

        else:
            cv = sols.get("cv")
            I, = np.where(cv == cv[cv.argmin()])

        return sols[I]


class MultiObjectiveArchive(Archive):

    def __new__(cls, max_size=200, truncate_size=100, **kwargs):
        return super().__new__(cls,
                               max_size=max_size,
                               truncate_size=truncate_size,
                               **kwargs)

    def _find_opt(self, sols):
        feas = sols.get("feas")

        if np.any(feas):
            sols = sols[feas]

            F = sols.get("F")
            I = NonDominatedSorting().do(F, only_non_dominated_front=True)
        else:
            cv = sols.get("cv")
            I, = np.where(cv == cv[cv.argmin()])

        return sols[I]


def default_archive(problem, **kwargs):
    if problem.n_obj == 1:
        return SingleObjectiveArchive(**kwargs)

    elif problem.n_obj == 2:
        from pymoo.algorithms.moo.sms import LeastHypervolumeContributionSurvival
        survival = LeastHypervolumeContributionSurvival()
        return MultiObjectiveArchive(truncation=SurvivalTruncation(survival, problem=problem), **kwargs)

    elif problem.n_obj >= 3:
        from pymoo.algorithms.moo.spea2 import SPEA2Survival
        survival = SPEA2Survival()
        return MultiObjectiveArchive(truncation=SurvivalTruncation(survival, problem=problem), **kwargs)



import numpy as np


from pymoo.util.misc import cdist

from pymoo.util.ref_dirs.misc import project_onto_unit_simplex_recursive
from pymoo.util.reference_direction import ReferenceDirectionFactory, sample_on_unit_simplex, \
    select_points_with_maximum_distance, get_partition_closest_to_points, UniformReferenceDirectionFactory


def kmeans(X, centroids, n_max_iter, a_tol, n_ignore):

    for i in range(n_max_iter):

        # copy the old centroids
        last_centroids = np.copy(centroids)

        # assign all points to one of the centroids
        points_to_centroid = cdist(X, centroids).argmin(axis=1)

        centroids_to_points = [[] for _ in range(len(centroids))]
        for j, k in enumerate(points_to_centroid):
            centroids_to_points[k].append(j)

        for j in range(n_ignore, len(centroids_to_points)):
            centroids[j] = np.mean(X[centroids_to_points[j]], axis=0)

        project_onto_unit_simplex_recursive(centroids)
        centroids /= centroids.sum(axis=1)[:, None]

        delta = np.abs(centroids - last_centroids).sum(axis=1).mean()

        if delta < a_tol:
            break

    return centroids


class ReductionBasedReferenceDirectionFactory(ReferenceDirectionFactory):

    def __init__(self,
                 n_dim,
                 n_points,
                 scaling=None,
                 n_sample_points=10000,
                 sampling="kraemer",
                 kmeans=True,
                 kmeans_max_iter=1000,
                 kmeans_a_tol=0.0001,
                 **kwargs):

        super().__init__(n_dim, scaling, **kwargs)
        self.n_sample_points = n_sample_points
        self.sampling = sampling
        self.kmeans = kmeans
        self.kmeans_max_iter = kmeans_max_iter
        self.kmeans_a_tol = kmeans_a_tol

        if n_points is None:
            raise Exception("Please provide the number of points to be factored!")

        self.n_points = n_points

    def _do(self):
        rnd = sample_on_unit_simplex(self.n_sample_points, self.n_dim, unit_simplex_mapping=self.sampling)

        def h(n):
            return get_partition_closest_to_points(n, self.n_dim)

        H = h(self.n_points)

        E = UniformReferenceDirectionFactory(self.n_dim, n_partitions=H).do()
        E = E[np.any(E == 0, axis=1)]

        # add the edge coordinates
        X = np.row_stack([E, rnd])

        I = select_points_with_maximum_distance(X, self.n_points, selected=list(range((len(E)))))
        centroids = X[I].copy()

        if self.kmeans:
            #centroids = kmeans(X, centroids, self.kmeans_max_iter, self.kmeans_a_tol, 0)
            centroids = kmeans(X, centroids, self.kmeans_max_iter, self.kmeans_a_tol, len(E))

        return centroids



import numpy as np

from pymoo.indicators.hv.exact import ExactHypervolume


def hvc_2d_slow(ref_point, F):
    n = len(F)

    I = np.lexsort((-F[:, 1], F[:, 0]))

    V = np.row_stack([ref_point, F[I], ref_point])

    hvi = np.zeros(n)

    for k in range(1, n + 1):
        height = V[k - 1, 1] - V[k, 1]
        width = V[k + 1, 0] - V[k, 0]

        hvi[I[k - 1]] = width * height

    return np.array(hvi)


def hvc_2d_fast(ref_point, F_sorted, left=None, right=None):
    if left is None:
        left = [F_sorted[0, 0], ref_point[1]]

    if right is None:
        right = [ref_point[0], F_sorted[-1, 1]]

    V = np.row_stack([left, F_sorted, right])
    height = (V[:-1, 1] - V[1:, 1])[:-1]
    width = (V[1:, 0] - V[:-1, 0])[1:]

    hvc = height * width
    return hvc


def hv_2d_fast(ref_point, F_sorted):
    V = np.row_stack([ref_point, F_sorted])
    height = (V[:-1, 1] - V[1:, 1])
    width = ref_point[0] - V[1:, 0]
    return (height * width).sum()


class ExactHypervolume2D(ExactHypervolume):

    def __init__(self, ref_point, **kwargs) -> None:
        assert len(ref_point) == 2, "This hypervolume calculation only works in 2 dimensions."
        super().__init__(ref_point, func_hv=hv_2d_fast, func_hvc=hvc_2d_fast, **kwargs)
        self.S = None
        self.I = None

    def _calc(self, ref_point, F):
        if len(F) == 0:
            self.I, self.S = [], []
            return 0.0, np.zeros(0)

        F = np.minimum(self.ref_point, self.F)
        I = np.lexsort((-F[:, 1], F[:, 0]))
        S = np.argsort(I)

        hv, hvc = super()._calc(ref_point, F[I])
        hvc = hvc[S]
        self.I, self.S = I, S

        return hv, hvc

    # this very efficiently just recomputes hvc of the points necessary
    # however has not shown to be much faster because of reindexing
    # def delete(self, k):
    #     assert k < len(self.F)
    #
    #     F, I, S, hv, hvc = self.F, self.I, self.S, self.hv, self.hvc
    #
    #     hv -= hvc[k]
    #
    #     i = S[k]
    #
    #     S = np.delete(S, k, axis=0)
    #     S[S > i] -= 1
    #
    #     I = np.delete(I, i, axis=0)
    #     I[I > k] -= 1
    #
    #     F = np.delete(F, k, axis=0)
    #     hvc = np.delete(self.hvc, k, axis=0)
    #
    #     v = [I[i] if 0 <= i < len(I) else None for i in np.arange(i-2, i+2)]
    #     left, middle, right = v[0], v[1:-1], v[-1]
    #
    #     middle = [e for e in middle if e is not None]
    #
    #     if len(middle) > 0:
    #
    #         hvc[middle] = hvc_2d_fast(self.ref_point,
    #                                   F[middle],
    #                                   left=F[left] if left is not None else None,
    #                                   right=F[right] if right is not None else None,
    #                                   )
    #
    #     self.F, self.I, self.S, self.hv, self.hvc = F, I, S, hv, hvc



import matplotlib.pyplot as plt
import numpy as np

from pymoo.visualization.fitness_landscape import FitnessLandscape
from pymoo.visualization.video.callback_video import AnimationCallback


class TwoVariablesOneObjectiveVisualization(AnimationCallback):

    def __init__(self,
                 n_samples_for_surface=10000,
                 **kwargs):
        super().__init__(**kwargs)
        self.last_pop = None
        self.n_samples_for_surface = n_samples_for_surface

    def do(self, problem, algorithm):

        # check whether the visualization can be done or not - throw exception or simply do nothing
        if problem.n_var != 2 or problem.n_obj != 1:
            raise Exception("This visualization can only be used for problems with two variables and one objective!")

        # draw the problem surface
        # if algorithm.surrogate.targets["F"].doe is not None:
        #     problem = algorithm.surrogate
        plot = FitnessLandscape(problem, _type="contour", kwargs_contour=dict(alpha=0.5))
        plot.do()

        # get the population
        pop = algorithm.pop

        X, F, CV = pop.get("X", "F", "CV")
        plt.scatter(X[:, 0], X[:, 1], facecolor="none", edgecolors="black", marker="o", s=50, label="Solutions")

        if hasattr(algorithm, "off") and algorithm.off is not None:
            X, F, CV = algorithm.off.get("X", "F", "CV")
            plt.scatter(X[:, 0], X[:, 1], color="green", marker="D", s=30, label="Offsprings")

        is_new = np.full(len(pop), True)
        if self.last_pop is not None:
            for k, ind in enumerate(pop):
                if ind in self.last_pop:
                    is_new[k] = False

        # plot the new population
        if is_new.sum() > 0:
            X, F, CV = pop[is_new].get("X", "F", "CV")
            plt.scatter(X[:, 0], X[:, 1], color="red", marker="*", s=70, label="Survivors")

        xl, xu = problem.bounds()
        plt.xlim(xl[0], xu[0])
        plt.ylim(xl[1], xu[1])

        plt.title(f"Generation: {algorithm.n_gen}")
        plt.legend()

        # store the current population as the last
        self.last_pop = set(pop)

        plt.show()

        return plt.gcf()



import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.core.population import Population
from pymoo.core.survival import Survival
from pymoo.docs import parse_doc_string
from pymoo.indicators.hv.exact import ExactHypervolume
from pymoo.indicators.hv.exact_2d import ExactHypervolume2D
from pymoo.indicators.hv.monte_carlo import ApproximateMonteCarloHypervolume
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.tournament import compare, TournamentSelection
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.dominator import Dominator
from pymoo.util.function_loader import load_function
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.util.normalization import normalize


# ---------------------------------------------------------------------------------------------------------
# Environmental Survival - Remove the solution with the least HV contribution
# ---------------------------------------------------------------------------------------------------------


class LeastHypervolumeContributionSurvival(Survival):

    def __init__(self, eps=10.0) -> None:
        super().__init__(filter_infeasible=True)
        self.eps = eps

    def _do(self, problem, pop, *args, n_survive=None, ideal=None, nadir=None, **kwargs):

        # get the objective space values and objects
        F = pop.get("F").astype(float, copy=False)

        # if the boundary points are not provided -> estimate them from pop
        if ideal is None:
            ideal = F.min(axis=0)
        if nadir is None:
            nadir = F.max(axis=0)

        # the number of objectives
        _, n_obj = F.shape

        # the final indices of surviving individuals
        survivors = []

        # do the non-dominated sorting until splitting front
        fronts = NonDominatedSorting().do(F, n_stop_if_ranked=n_survive)

        for k, front in enumerate(fronts):

            # get the actual front as individuals
            front = pop[front]
            front.set("rank", k)

            if len(survivors) + len(front) > n_survive:

                # normalize all the function values for the front
                F = front.get("F")
                F = normalize(F, ideal, nadir)

                # define the reference point and shift it a bit - F is already normalized!
                ref_point = np.full(problem.n_obj, 1.0 + self.eps)

                _, n_obj = F.shape

                # choose the suitable hypervolume method
                clazz = ExactHypervolume
                if n_obj == 2:
                    clazz = ExactHypervolume2D
                elif n_obj > 3:
                    clazz = ApproximateMonteCarloHypervolume

                # finally do the computation
                hv = clazz(ref_point).add(F)

                # current front sorted by crowding distance if splitting
                while len(survivors) + len(front) > n_survive:
                    k = hv.hvc.argmin()
                    hv.delete(k)
                    front = np.delete(front, k)

            # extend the survivors by all or selected individuals
            survivors.extend(front)

        return Population.create(*survivors)


# ---------------------------------------------------------------------------------------------------------
# Binary Tournament
# ---------------------------------------------------------------------------------------------------------


def cv_and_dom_tournament(pop, P, *args, **kwargs):
    n_tournaments, n_parents = P.shape

    if n_parents != 2:
        raise ValueError("Only implemented for binary tournament!")

    S = np.full(n_tournaments, np.nan)

    for i in range(n_tournaments):

        a, b = P[i, 0], P[i, 1]
        a_cv, a_f, b_cv, b_f, = pop[a].CV[0], pop[a].F, pop[b].CV[0], pop[b].F

        # if at least one solution is infeasible
        if a_cv > 0.0 or b_cv > 0.0:
            S[i] = compare(a, a_cv, b, b_cv, method='smaller_is_better', return_random_if_equal=True)

        # both solutions are feasible
        else:

            # if one dominates another choose the nds one
            rel = Dominator.get_relation(a_f, b_f)
            if rel == 1:
                S[i] = a
            elif rel == -1:
                S[i] = b

            # if rank or domination relation didn't make a decision compare by crowding
            if np.isnan(S[i]):
                S[i] = np.random.choice([a, b])

    return S[:, None].astype(int, copy=False)


# ---------------------------------------------------------------------------------------------------------
# Algorithm
# ---------------------------------------------------------------------------------------------------------

class SMSEMOA(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=FloatRandomSampling(),
                 selection=TournamentSelection(func_comp=cv_and_dom_tournament),
                 crossover=SBX(prob_exch=0.5),
                 mutation=PM(),
                 survival=LeastHypervolumeContributionSurvival(),
                 eliminate_duplicates=True,
                 n_offsprings=None,
                 normalize=True,
                 output=MultiObjectiveOutput(),
                 **kwargs):
        """

        Parameters
        ----------
        pop_size : {pop_size}
        sampling : {sampling}
        selection : {selection}
        crossover : {crossover}
        mutation : {mutation}
        eliminate_duplicates : {eliminate_duplicates}
        n_offsprings : {n_offsprings}

        """
        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         survival=survival,
                         eliminate_duplicates=eliminate_duplicates,
                         n_offsprings=n_offsprings,
                         output=output,
                         advance_after_initial_infill=True,
                         **kwargs)

        self.normalize = normalize

    def _advance(self, infills=None, **kwargs):

        ideal, nadir = None, None

        # estimate ideal and nadir from the current population (more robust then from doing it from merged)
        if self.normalize:
            F = self.pop.get("F")
            ideal, nadir = F.min(axis=0), F.max(axis=0) + 1e-32

        # merge the offsprings with the current population
        if infills is not None:
            pop = Population.merge(self.pop, infills)

        self.pop = self.survival.do(self.problem, pop, n_survive=self.pop_size, algorithm=self,
                                    ideal=ideal, nadir=nadir, **kwargs)


parse_doc_string(SMSEMOA.__init__)



from pymoo.core.variable import Variable


# ---------------------------------------------------------------------------------------------------------
# Functions
# ---------------------------------------------------------------------------------------------------------

def get_data(obj):
    if not isinstance(obj, dict):
        if hasattr(obj, "__dict__"):
            return obj.__dict__
        else:
            return {}
    else:
        return obj


def get_params(obj, flag="default", only_active=True):
    return get_params_bfs(obj, flag, only_active)


def get_params_bfs(obj, flag, only_active):
    ret = {}

    q = [(None, obj)]

    visited = set()

    while len(q) > 0:

        prefix, obj = q.pop()

        if isinstance(obj, Variable):

            if obj not in visited and obj.flag == flag and (not only_active or obj.active):

                # find the right spot in the ret dictionary
                e = ret

                for name in prefix[:-1]:
                    if name not in e:
                        e[name] = {}
                    e = e[name]

                # set it so the value
                e[prefix[-1]] = obj

            # add it to have been visited
            visited.add(obj)

        else:
            data = get_data(obj)
            for key in data:
                new_prefix = [key] if prefix is None else prefix + [key]
                entry = (new_prefix, data[key])
                q.append(entry)

    return ret


def get_params_rec(obj, visited, flag, only_active):
    data = get_data(obj)

    ret = {}
    for k, v in data.items():
        if isinstance(v, Variable):

            if v not in visited and v.flag == flag and (not only_active or v.active):
                ret[k] = v

            visited.add(v)

        else:
            entry = get_params_rec(v, visited, flag, only_active)
            if entry is not None and len(entry) > 0:
                ret[k] = entry
    return ret


def apply_to_params(obj, func_apply):
    for _, v in flatten_rec(get_params(obj)):
        func_apply(v)


def deactivate_params(obj):
    def func(param):
        param.active = False

    apply_to_params(obj, func)


def set_params(obj, params, as_value=True):
    data = get_data(obj)

    for k, v in params.items():
        if isinstance(v, dict):
            set_params(data[k], v)
        else:
            if as_value:
                data[k].set(v)
            else:
                data[k] = v


def flatten(params):
    return {k: v for k, v in flatten_rec(params)}


def flatten_rec(params, prefix=None):
    if hasattr(params, "items"):
        for k, v in params.items():
            yield from flatten_rec(v, prefix=f"{prefix}.{k}" if prefix is not None else k)
    else:
        yield prefix, params


def hierarchical(data):
    ret = {}

    groups = {}
    for k, v in data.items():
        a = k.split(".")
        if len(a) > 1:
            prefix = a[0]
            if prefix not in groups:
                groups[prefix] = {}
            groups[prefix][".".join(a[1:])] = v
        else:
            ret[k] = v

    for name, group in groups.items():
        ret[name] = hierarchical(group)

    return ret



import numpy as np

from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.docs import parse_doc_string
from pymoo.core.survival import Survival
from pymoo.operators.selection.rnd import RandomSelection
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.util.normalization import get_extreme_points_c

# =========================================================================================================
# Implementation
# =========================================================================================================


class RNSGA2(NSGA2):

    def __init__(self,
                 ref_points,
                 epsilon=0.001,
                 normalization="front",
                 weights=None,
                 extreme_points_as_reference_points=False,
                 **kwargs):
        """

        Parameters
        ----------

        ref_points : {ref_points}
        epsilon : float
        weights : np.array
        normalization : {{'no', 'front', 'ever'}}
        extreme_points_as_reference_points : bool

        """

        self.epsilon = epsilon
        self.weights = weights
        self.normalization = normalization
        self.selection = RandomSelection()

        super().__init__(**kwargs)

        self.survival = RankAndModifiedCrowdingSurvival(ref_points, epsilon, weights, normalization,
                                                        extreme_points_as_reference_points)


class RankAndModifiedCrowdingSurvival(Survival):

    def __init__(self, ref_points,
                 epsilon,
                 weights,
                 normalization,
                 extreme_points_as_reference_points
                 ) -> None:

        super().__init__(True)
        self.n_obj = ref_points.shape[1]
        self.ref_points = ref_points
        self.epsilon = epsilon
        self.extreme_points_as_reference_points = extreme_points_as_reference_points

        self.weights = weights
        if self.weights is None:
            self.weights = np.full(self.n_obj, 1 / self.n_obj)

        self.normalization = normalization
        self.ideal_point = np.full(self.n_obj, np.inf)
        self.nadir_point = np.full(self.n_obj, -np.inf)

    def _do(self, problem, pop, n_survive=None, **kwargs):

        # get the objective space values and objects
        F = pop.get("F")

        # the final indices of surviving individuals
        survivors = []

        # do the non-dominated sorting until splitting front
        fronts = NonDominatedSorting().do(F)

        if self.normalization == "ever":
            # find or usually update the new ideal point - from feasible solutions
            self.ideal_point = np.min(np.vstack((self.ideal_point, F)), axis=0)
            self.nadir_point = np.max(np.vstack((self.nadir_point, F)), axis=0)

        elif self.normalization == "front":
            front = fronts[0]
            if len(front) > 1:
                self.ideal_point = np.min(F[front], axis=0)
                self.nadir_point = np.max(F[front], axis=0)

        elif self.normalization == "no":
            self.ideal_point = np.zeros(self.n_obj)
            self.nadir_point = np.ones(self.n_obj)

        if self.extreme_points_as_reference_points:
            self.ref_points = np.row_stack([self.ref_points, get_extreme_points_c(F, self.ideal_point)])

        # calculate the distance matrix from ever solution to all reference point
        dist_to_ref_points = calc_norm_pref_distance(F, self.ref_points, self.weights, self.ideal_point,
                                                     self.nadir_point)

        for k, front in enumerate(fronts):

            # save rank attributes to the individuals - rank = front here
            pop[front].set("rank", np.full(len(front), k))

            # number of individuals remaining
            n_remaining = n_survive - len(survivors)

            # the ranking of each point regarding each reference point (two times argsort is necessary)
            rank_by_distance = np.argsort(np.argsort(dist_to_ref_points[front], axis=0), axis=0)

            # the reference point where the best ranking is coming from
            ref_point_of_best_rank = np.argmin(rank_by_distance, axis=1)

            # the actual ranking which is used as crowding
            ranking = rank_by_distance[np.arange(len(front)), ref_point_of_best_rank]

            if len(front) <= n_remaining:

                # we can simply copy the crowding to ranking. not epsilon selection here
                crowding = ranking
                I = np.arange(len(front))

            else:

                # Distance from solution to every other solution and set distance to itself to infinity
                dist_to_others = calc_norm_pref_distance(F[front], F[front], self.weights, self.ideal_point,
                                                         self.nadir_point)
                np.fill_diagonal(dist_to_others, np.inf)

                # the crowding that will be used for selection
                crowding = np.full(len(front), np.nan)

                # solutions which are not already selected - for
                not_selected = np.argsort(ranking)

                # until we have saved a crowding for each solution
                while len(not_selected) > 0:

                    # select the closest solution
                    idx = not_selected[0]

                    # set crowding for that individual
                    crowding[idx] = ranking[idx]

                    # need to remove myself from not-selected array
                    to_remove = [idx]

                    # Group of close solutions
                    dist = dist_to_others[idx][not_selected]
                    group = not_selected[np.where(dist < self.epsilon)[0]]

                    # if there exists solution with a distance less than epsilon
                    if len(group):
                        # discourage them by giving them a high crowding
                        crowding[group] = ranking[group] + np.round(len(front) / 2)

                        # remove group from not_selected array
                        to_remove.extend(group)

                    not_selected = np.array([i for i in not_selected if i not in to_remove])

                # now sort by the crowding (actually modified rank) ascending and let the best survive
                I = np.argsort(crowding)[:n_remaining]

            # set the crowding to all individuals
            pop[front].set("crowding", crowding)

            # extend the survivors by all or selected individuals
            survivors.extend(front[I])

        # inverse of crowding because nsga2 does maximize it (then tournament selection can stay the same)
        pop.set("crowding", -pop.get("crowding"))

        return pop[survivors]


def calc_norm_pref_distance(A, B, weights, ideal, nadir):
    D = np.repeat(A, B.shape[0], axis=0) - np.tile(B, (A.shape[0], 1))
    N = ((D / (nadir - ideal)) ** 2) * weights
    N = np.sqrt(np.sum(N, axis=1) * len(weights))
    return np.reshape(N, (A.shape[0], B.shape[0]))


parse_doc_string(RNSGA2.__init__)



import numpy as np

from pymoo.core.repair import Repair


class RoundingRepair(Repair):

    def __init__(self, **kwargs) -> None:
        """

        Returns
        -------
        object
        """
        super().__init__(**kwargs)

    def _do(self, problem, X, **kwargs):
        return np.around(X).astype(int)



from pymoo.core.decomposition import Decomposition


class ASF(Decomposition):

    def _do(self, F, weights, weight_0=1e-10, **kwargs):
        _weights = weights.astype(float)
        _weights[weights == 0] = weight_0
        asf = ((F - self.utopian_point) / _weights).max(axis=1)
        return asf



import argparse
import copy
import distutils
import os
import sys
import traceback

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext

from pymoo.version import __version__

# ---------------------------------------------------------------------------------------------------------
# GENERAL
# ---------------------------------------------------------------------------------------------------------


__name__ = "pymoo"
__author__ = "Julian Blank"
__url__ = "https://pymoo.org"
data = dict(
    name=__name__,
    version=__version__,
    author=__author__,
    url=__url__,
    python_requires='>=3.7',
    author_email="blankjul@msu.edu",
    description="Multi-Objective Optimization in Python",
    license='Apache License 2.0',
    keywords="optimization",
    # packages=["pymoo"] + ["pymoo." + e for e in find_packages(where='pymoo')],
    packages=find_packages(include=['pymoo', 'pymoo.*']),
    include_package_data=True,
    exclude_package_data={
        '': ['*.c', '*.cpp', '*.pyx'],
    },
    install_requires=['numpy>=1.15',
                      'scipy>=1.1',
                      'matplotlib>=3',
                      'autograd>=1.4',
                      'cma==3.2.2',
                      'alive-progress',
                      'dill',
                      'Deprecated'],
    platforms='any',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Mathematics'
    ]
)


# ---------------------------------------------------------------------------------------------------------
# OTHER METADATA
# ---------------------------------------------------------------------------------------------------------


# update the readme.rst to be part of setup
def readme():
    with open('README.rst') as f:
        return f.read()


data['long_description'] = readme()
data['long_description_content_type'] = 'text/x-rst'

# ---------------------------------------------------------------------------------------------------------
# OPTIONS
# ---------------------------------------------------------------------------------------------------------


parser = argparse.ArgumentParser()
parser.add_argument('--nopyx', dest='nopyx', action='store_true',
                    help='Whether the pyx files shall be considered at all.')
parser.add_argument('--nocython', dest='nocython', action='store_true', help='Whether pyx files shall be cythonized.')
parser.add_argument('--nolibs', dest='nolibs', action='store_true', help='Whether the libraries should be compiled.')
args, _ = parser.parse_known_args()

sys.argv = [e for e in sys.argv if not e.lstrip("-") in args]


# ============================================================
# MacOSX FIX for compiling modules
# ============================================================

def is_new_osx():
    name = distutils.util.get_platform()
    if sys.platform != "darwin":
        return False
    elif name.startswith("macosx-10"):
        minor_version = int(name.split("-")[1].split(".")[1])
        if minor_version >= 7:
            return True
        else:
            return False
    else:
        return False


# fix compiling for new macosx!
if is_new_osx():
    os.environ['CFLAGS'] = '-stdlib=libc++'


# ============================================================
# Module for Compilation - Throws an Exception if Failing
# ============================================================


# exception that is thrown when the build fails
class CompilingFailed(Exception):
    pass


# try to compile, if not possible throw exception
def construct_build_ext(build_ext):
    class WrappedBuildExt(build_ext):
        def run(self):
            try:
                build_ext.run(self)
            except BaseException as e:
                raise CompilingFailed(e)

        def build_extension(self, ext):
            try:
                build_ext.build_extension(self, ext)
            except BaseException as e:
                raise CompilingFailed(e)

    return WrappedBuildExt


# ============================================================
# SETUP
# ============================================================

ROOT = os.path.dirname(os.path.realpath(__file__))
backup = copy.deepcopy(data)

try:

    if args.nopyx:
        ext = []
    elif args.nocython:
        path = os.path.join(ROOT, "pymoo", "cython")
        pyx = [os.path.join(path, f) for f in os.listdir() if f.endswith(".pyx")]
        ext = [Extension(f"pymoo.cython.{source[:-4]}", [source]) for source in pyx]
    else:
        from Cython.Build import cythonize

        ext = cythonize("pymoo/cython/*.pyx")

    if not args.nolibs:

        if len(ext) > 0:
            data['ext_modules'] = ext

            try:
                import numpy as np

                data['include_dirs'] = [np.get_include()]
            except BaseException:
                raise CompilingFailed(
                    "NumPy libraries must be installed for compiled extensions! Speedups are not enabled.")

            # return the object for building which allows installation with no compilation
            data['cmdclass'] = dict(build_ext=construct_build_ext(build_ext))

    setup(**data)
    print('*' * 75)
    print("Compilation Successful.")
    print("Installation with compiled libraries succeeded.")
    print('*' * 75)

except BaseException:

    # retrieve the original input arguments and execute default setup
    setup(**backup)

    # get information why compiling has failed
    ex_type, ex_value, ex_traceback = sys.exc_info()

    print('*' * 75)
    print("WARNING: Compilation Failed.")
    print("WARNING:", ex_type)
    print("WARNING:", ex_value)
    print()
    print("=" * 75)
    traceback.print_exc()
    print("=" * 75)
    print()
    print("WARNING: For the compiled libraries numpy is required. Please make sure they are installed")
    print("WARNING: pip install numpy")
    print("WARNING: Also, make sure you have a compiler for C++!")

    print('*' * 75)
    print("Plain Python installation succeeded.")
    print('*' * 75)



import numpy as np

from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival, GA
from pymoo.core.survival import Survival
from pymoo.docs import parse_doc_string
from pymoo.operators.selection.tournament import compare, TournamentSelection
from pymoo.termination.cv import ConstraintViolationTermination
from pymoo.termination.default import DefaultSingleObjectiveTermination, DefaultTermination
from pymoo.termination.ftol import SingleObjectiveSpaceTermination
from pymoo.termination.robust import RobustTermination
from pymoo.termination.xtol import DesignSpaceTermination
from pymoo.util.clearing import EpsilonClearing
from pymoo.util.display.column import Column
from pymoo.util.display.single import SingleObjectiveOutput
from pymoo.util.misc import norm_eucl_dist


# =========================================================================================================
# Display
# =========================================================================================================

class NicheOutput(SingleObjectiveOutput):

    def __init__(self):
        super().__init__()
        self.n_niches = Column("n_niches", width=10, func=lambda algorithm: len(algorithm.opt))
        self.columns += [self.n_niches]


# =========================================================================================================
# Termination
# =========================================================================================================

class NicheSingleObjectiveSpaceToleranceTermination(SingleObjectiveSpaceTermination):

    def _data(self, algorithm):
        return algorithm.opt.get("F").mean()


class NicheTermination(DefaultTermination):

    def __init__(self,
                 x_tol=1e-32,
                 cv_tol=1e-6,
                 f_tol=1e-6,
                 period=20,
                 **kwargs) -> None:
        super().__init__(RobustTermination(DesignSpaceTermination(tol=x_tol), period=period),
                         RobustTermination(ConstraintViolationTermination(tol=cv_tol), period=period),
                         RobustTermination(NicheSingleObjectiveSpaceToleranceTermination(tol=f_tol, n_skip=5),
                                           period=period),
                         **kwargs)


# =========================================================================================================
# Selection
# =========================================================================================================


def comp_by_cv_and_clearing_fitness(pop, P, **kwargs):
    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):
        a, b = P[i, 0], P[i, 1]

        # if at least one solution is infeasible
        if pop[a].CV[0] > 0.0 or pop[b].CV[0] > 0.0:
            S[i] = compare(a, pop[a].CV, b, pop[b].CV,
                           method='smaller_is_better',
                           return_random_if_equal=True)

        # first compare by the round the individual was selected
        else:
            S[i] = compare(a, pop[a].get("iter"), b, pop[b].get("iter"), method='smaller_is_better')

            # if it was the same round - then use the rank of the fitness directly
            if np.isnan(S[i]):
                S[i] = compare(a, pop[a].get("rank"), b, pop[b].get("rank"),
                               method='smaller_is_better', return_random_if_equal=True)

    return S[:, None].astype(int)


# =========================================================================================================
# Survival
# =========================================================================================================

class EpsilonClearingSurvival(Survival):

    def __init__(self, epsilon, n_max_each_iter=None, norm_by_dim=False) -> None:
        super().__init__(False)
        self.epsilon = epsilon
        self.n_max_each_iter = n_max_each_iter
        self.norm_by_dim = norm_by_dim

    def _do(self, problem, pop, n_survive=None, out=None, **kwargs):
        F = pop.get("F")

        if F.shape[1] != 1:
            raise ValueError("FitnessSurvival can only used for single objective single!")

        # this basically sorts the population by constraint and objective value
        pop = FitnessSurvival().do(problem, pop, n_survive=len(pop))

        # calculate the distance from each individual to another - pre-processing for the clearing
        # NOTE: the distance is normalized by the maximum distance possible
        X = pop.get("X").astype(float)
        D = norm_eucl_dist(problem, X, X)
        if self.norm_by_dim:
            D = D / (problem.n_var ** 0.5)

        # initialize the clearing strategy
        clearing = EpsilonClearing(D, self.epsilon)

        # initialize the iteration and rank i the beginning
        iter, rank = 1, 1

        # also solutions that have been found in the first iteration
        iter_one = None

        # until the number of selected individuals are less than expected survivors
        while len(clearing.selected()) < n_survive:

            # get all the remaining indices
            remaining = clearing.remaining()

            # if no individuals are left because of clearing - perform a reset
            if len(remaining) == 0 or (self.n_max_each_iter is not None and rank > self.n_max_each_iter):
                # reset and retrieve the newly available indices
                clearing.reset()
                remaining = clearing.remaining()

                # increase the iteration counter and start over from rank 1
                iter += 1
                rank = 1

                # get the individual of the first iteration - needed for niche assignment
                iter_one = np.where(pop.get("iter") == 1)[0] if iter_one is None else iter_one

            # since the population is ordered by F and CV it is always the first index
            k = remaining[0]

            # set the attribute to the selected individual
            pop[k].set("iter", iter)
            pop[k].set("rank", rank)

            # in the first iteration set the niche counter for each solution equal to rank
            if iter == 1:
                pop[k].set("niche", rank)
            else:
                closest_iter_one = iter_one[D[k][iter_one].argmin()]
                niche = pop[closest_iter_one].get("niche")
                pop[k].set("niche", niche)

            clearing.select(k)
            rank += 1

        # retrieve all individuals being selected
        S = clearing.selected()

        return pop[S]


# =========================================================================================================
# Algorithm
# =========================================================================================================


class NicheGA(GA):

    def __init__(self,
                 pop_size=100,
                 norm_niche_size=0.05,
                 norm_by_dim=False,
                 return_all_opt=True,
                 output=NicheOutput(),
                 **kwargs):
        """

        Parameters
        ----------
        norm_niche_size : float
            The radius in which the clearing shall be performed. The clearing is performed in the normalized design
            space, e.g. 0.05 corresponds to clear all solutions which have less norm euclidean distance than 5%.
        pop_size : {pop_size}
        sampling : {sampling}
        selection : {selection}
        crossover : {crossover}
        mutation : {mutation}
        eliminate_duplicates : {eliminate_duplicates}
        n_offsprings : {n_offsprings}

        """

        surv = kwargs.get("survival")
        if surv is None:
            surv = EpsilonClearingSurvival(norm_niche_size, n_max_each_iter=None, norm_by_dim=norm_by_dim)

        selection = kwargs.get("selection")
        if selection is None:
            selection = TournamentSelection(comp_by_cv_and_clearing_fitness)

        super().__init__(pop_size=pop_size,
                         selection=selection,
                         survival=surv,
                         output=output,
                         advance_after_initial_infill=True,
                         **kwargs)

        # self.termination = NicheTermination()
        self.termination = DefaultSingleObjectiveTermination()

        # whether with rank one after clearing or just the best should be considered as optimal
        self.return_all_opt = return_all_opt

    def _set_optimum(self, **kwargs):
        if self.return_all_opt:
            self.opt = self.pop[self.pop.get("iter") == 1]
        else:
            super()._set_optimum()


parse_doc_string(NicheGA.__init__)



import numpy as np

from pymoo.core.operator import Operator


class Repair(Operator):

    def do(self, problem, pop, **kwargs):
        X = np.array([ind.X for ind in pop])
        if self.vtype is not None:
            X = X.astype(self.vtype)

        Xp = self._do(problem, X, **kwargs)

        pop.set("X", Xp)
        return pop

    def _do(self, problem, X, **kwargs):
        return X


class NoRepair(Repair):
    pass




import numpy as np

from pymoo.core.problem import ElementwiseProblem


def func_return_none(*args, **kwargs):
    return None


class FunctionalProblem(ElementwiseProblem):

    def __init__(self,
                 n_var,
                 objs,
                 constr_ieq=[],
                 constr_eq=[],
                 func_pf=func_return_none,
                 func_ps=func_return_none,
                 **kwargs):

        # if only a single callable is provided (for single-objective problems) convert it to a list
        if callable(objs):
            objs = [objs]

        self.objs = objs
        self.constr_ieq = constr_ieq
        self.constr_eq = constr_eq
        self.func_pf = func_pf
        self.func_ps = func_ps

        super().__init__(n_var=n_var,
                         n_obj=len(self.objs),
                         n_ieq_constr=len(constr_ieq),
                         n_eq_constr=len(constr_eq),
                         **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = np.array([obj(x) for obj in self.objs])
        out["G"] = np.array([constr(x) for constr in self.constr_ieq])
        out["H"] = np.array([constr(x) for constr in self.constr_eq])

    def _calc_pareto_front(self, *args, **kwargs):
        return self.func_pf(*args, **kwargs)

    def _calc_pareto_set(self, *args, **kwargs):
        return self.func_ps(*args, **kwargs)




from pymoo.core.repair import Repair


class TypeRepair(Repair):

    def __init__(self, vtype) -> None:
        super().__init__()
        self.vtype = vtype

    def _do(self, problem, X, **kwargs):
        return X.astype(self.vtype)



import numpy as np

from pymoo.algorithms.base.line import LineSearchProblem
from pymoo.core.algorithm import Algorithm
from pymoo.core.evaluator import Evaluator
from pymoo.core.individual import Individual
from pymoo.core.solution import SolutionSet


class WolfeSearch(Algorithm):

    def __init__(self, c1=1e-4, c2=0.9, max_iter=10, **kwargs):

        super().__init__(**kwargs)
        self.c1 = c1
        self.c2 = c2
        self.max_iter = max_iter

    def _setup(self, problem, **kwargs):
        assert isinstance(problem,
                          LineSearchProblem), "The wolfe search only purpose is to solve a line search problem!"
        self.pop = SolutionSet.create(problem.point)
        self.opt = self.pop

    def _set_optimum(self, force=False):
        pass

    def step(self):
        sol = self._infill()

        self.opt = sol
        self.termination.force_termination = True

    def _infill(self):

        problem, evaluator = self.problem, self.evaluator
        evaluator.skip_already_evaluated = False

        sol, direction = self.problem.point, self.problem.direction

        # the function value and gradient of the initial solution
        sol.set("alpha", 0.0)
        sol_F, sol_dF = sol.F[0], sol.get("dF")[0]

        def zoom(alpha_low, alpha_high, max_iter=100):

            while True:

                _alpha = (alpha_high.get("alpha") + alpha_low.get("alpha")) / 2
                _point = Individual(X=_alpha)
                evaluator.eval(problem, _point, evaluate_values_of=["F", "CV"])

                if _point.F[0] > sol_F + self.c1 * _alpha * sol_dF @ direction or _point.F[0] > alpha_low.F[0]:
                    alpha_high = _point
                else:
                    evaluator.eval(problem, _point, evaluate_values_of=["dF"])
                    point_dF = _point.get("dF")[0]

                    if np.abs(point_dF @ direction) <= -self.c2 * sol_dF @ direction:
                        return _point

                    if (point_dF @ direction) * (alpha_high.get("alpha") - alpha_low.get("alpha")) >= 0:
                        alpha_high = alpha_low

                    alpha_low = _point

        last = sol

        alpha = 1.0
        current = Individual(X=alpha)

        for i in range(1, self.max_iter + 1):

            # evaluate the solutions
            evaluator.eval(problem, current, evaluate_values_of=["F", "CV"])

            # get the values from the solution to be used to evaluate the conditions
            F, dF, _F = last.F[0], last.get("dF")[0], current.F[0]

            # if the wolfe condition is violate we have found our upper bound
            if _F > sol_F + self.c1 * sol_dF @ direction or (i > 1 and F >= _F):
                return zoom(last, current)

            # for the other condition we need the gradient information
            evaluator.eval(problem, current, evaluate_values_of=["dF"])
            _dF = current.get("dF")[0]

            if np.abs(_dF @ direction) <= -self.c2 * sol_dF @ direction:
                return current

            if _dF @ direction >= 0:
                return zoom(current, last)

            alpha = 2 * alpha
            last = current
            current = Individual(X=alpha)

        return current


def wolfe_line_search(problem, sol, direction, c1=1e-4, c2=0.9, max_iter=10, evaluator=None):
    # initialize the evaluator to be used (this will make sure evaluations are counted)
    evaluator = evaluator if evaluator is not None else Evaluator()
    evaluator.skip_already_evaluated = False

    # the function value and gradient of the initial solution
    sol.set("alpha", 0.0)
    sol_F, sol_dF = sol.F[0], sol.get("dF")[0]

    def zoom(alpha_low, alpha_high, max_iter=100):

        while True:

            _alpha = (alpha_high.get("alpha") + alpha_low.get("alpha")) / 2
            _point = Individual(X=sol.X + _alpha * direction, alpha=_alpha)
            evaluator.eval(problem, _point, evaluate_values_of=["F", "CV"])

            if _point.F[0] > sol_F + c1 * _alpha * sol_dF @ direction or _point.F[0] > alpha_low.F[0]:
                alpha_high = _point
            else:
                evaluator.eval(problem, _point, evaluate_values_of=["dF"])
                point_dF = _point.get("dF")[0]

                if np.abs(point_dF @ direction) <= -c2 * sol_dF @ direction:
                    return _point

                if (point_dF @ direction) * (alpha_high.get("alpha") - alpha_low.get("alpha")) >= 0:
                    alpha_high = alpha_low

                alpha_low = _point

    last = sol

    alpha = 1.0
    current = Individual(X=sol.X + alpha * direction, alpha=alpha)

    for i in range(1, max_iter + 1):

        # evaluate the solutions
        evaluator.eval(problem, current, evaluate_values_of=["F", "CV"])

        # get the values from the solution to be used to evaluate the conditions
        F, dF, _F = last.F[0], last.get("dF")[0], current.F[0]

        # if the wolfe condition is violate we have found our upper bound
        if _F > sol_F + c1 * sol_dF @ direction or (i > 1 and F >= _F):
            return zoom(last, current)

        # for the other condition we need the gradient information
        evaluator.eval(problem, current, evaluate_values_of=["dF"])
        _dF = current.get("dF")[0]

        if np.abs(_dF @ direction) <= -c2 * sol_dF @ direction:
            return current

        if _dF @ direction >= 0:
            return zoom(current, last)

        alpha = 2 * alpha
        last = current
        current = Individual(X=sol.X + alpha * direction, alpha=alpha)

    return current



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.individual import calc_cv
from pymoo.core.meta import Meta
from pymoo.core.problem import Problem
from pymoo.util.misc import from_dict


class ConstraintsAsObjective(Meta, Problem):

    def __init__(self,
                 problem,
                 config=None,
                 append=True):

        super().__init__(problem)
        self.config = config
        self.append = append

        if append:
            self.n_obj = problem.n_obj + 1
        else:
            self.n_obj = 1

        self.n_ieq_constr = 0
        self.n_eq_constr = 0

    def do(self, X, return_values_of, *args, **kwargs):
        out = self.__object__.do(X, return_values_of, *args, **kwargs)

        # get at the values from the output
        F, G, H = from_dict(out, "F", "G", "H")

        # store a backup of the values in out
        out["__F__"], out["__G__"], out["__H__"] = F, G, H

        # calculate the total constraint violation (here normalization shall be already included)
        CV = calc_cv(G=G, H=H, config=self.config)

        # append the constraint violation as objective
        if self.append:
            out["F"] = anp.column_stack([CV, F])
        else:
            out["F"] = CV

        del out["G"]
        del out["H"]

        return out

    def pareto_front(self, *args, **kwargs):
        pf = super().pareto_front(*args, **kwargs)
        if pf is not None:
            pf = np.column_stack([np.zeros(len(pf)), pf])
        return pf



import math

import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.core.duplicate import ElementwiseDuplicateElimination
from pymoo.core.individual import Individual
from pymoo.core.infill import InfillCriterion
from pymoo.core.population import Population
from pymoo.core.problem import Problem
from pymoo.core.sampling import Sampling
from pymoo.core.variable import Choice, Real, Integer, Binary
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.crossover.ux import UX
from pymoo.operators.mutation.bitflip import BFM
from pymoo.operators.mutation.pm import PM
from pymoo.operators.mutation.rm import ChoiceRandomMutation
from pymoo.operators.repair.rounding import RoundingRepair
from pymoo.operators.selection.rnd import RandomSelection
from pymoo.util.display.single import SingleObjectiveOutput


class MixedVariableMating(InfillCriterion):

    def __init__(self,
                 selection=RandomSelection(),
                 crossover=None,
                 mutation=None,
                 repair=None,
                 eliminate_duplicates=True,
                 n_max_iterations=100,
                 **kwargs):

        super().__init__(repair, eliminate_duplicates, n_max_iterations, **kwargs)

        if crossover is None:
            crossover = {
                Binary: UX(),
                Real: SBX(),
                Integer: SBX(vtype=float, repair=RoundingRepair()),
                Choice: UX(),
            }

        if mutation is None:
            mutation = {
                Binary: BFM(),
                Real: PM(),
                Integer: PM(vtype=float, repair=RoundingRepair()),
                Choice: ChoiceRandomMutation(),
            }

        self.selection = selection
        self.crossover = crossover
        self.mutation = mutation

    def _do(self, problem, pop, n_offsprings, parents=False, **kwargs):

        # So far we assume all crossover need the same amount of parents and create the same number of offsprings
        XOVER_N_PARENTS = 2
        XOVER_N_OFFSPRINGS = 2

        # the variables with the concrete information
        vars = problem.vars

        # group all the variables by their types
        vars_by_type = {}
        for k, v in vars.items():
            clazz = type(v)

            if clazz not in vars_by_type:
                vars_by_type[clazz] = []
            vars_by_type[clazz].append(k)

        # # all different recombinations (the choices need to be split because of data types)
        recomb = []
        for clazz, list_of_vars in vars_by_type.items():
            if clazz == Choice:
                for e in list_of_vars:
                    recomb.append((clazz, [e]))
            else:
                recomb.append((clazz, list_of_vars))

        # create an empty population that will be set in each iteration
        off = Population.new(X=[{} for _ in range(n_offsprings)])

        if not parents:
            n_select = math.ceil(n_offsprings / XOVER_N_OFFSPRINGS)
            pop = self.selection(problem, pop, n_select, XOVER_N_PARENTS, **kwargs)

        for clazz, list_of_vars in recomb:

            crossover = self.crossover[clazz]
            assert crossover.n_parents == XOVER_N_PARENTS and crossover.n_offsprings == XOVER_N_OFFSPRINGS

            _parents = [[Individual(X=np.array([parent.X[var] for var in list_of_vars])) for parent in parents] for
                        parents in pop]

            _vars = [vars[e] for e in list_of_vars]
            _xl, _xu = None, None

            if clazz in [Real, Integer]:
                _xl, _xu = np.array([v.bounds for v in _vars]).T

            _problem = Problem(vars=_vars, xl=_xl, xu=_xu)

            _off = crossover(_problem, _parents, **kwargs)

            mutation = self.mutation[clazz]
            _off = mutation(_problem, _off, **kwargs)

            for k in range(n_offsprings):
                for i, name in enumerate(list_of_vars):
                    off[k].X[name] = _off[k].X[i]

        return off


class MixedVariableSampling(Sampling):

    def _do(self, problem, n_samples, **kwargs):
        V = {name: var.sample(n_samples) for name, var in problem.vars.items()}

        X = []
        for k in range(n_samples):
            X.append({name: V[name][k] for name in problem.vars.keys()})

        return X


class MixedVariableDuplicateElimination(ElementwiseDuplicateElimination):

    def is_equal(self, a, b):
        a, b = a.X, b.X
        for k, v in a.items():
            if k not in b or b[k] != v:
                return False
        return True


def groups_of_vars(vars):
    ret = {}
    for name, var in vars.items():
        if var.__class__ not in ret:
            ret[var.__class__] = []

        ret[var.__class__].append((name, var))

    return ret


class MixedVariableGA(GeneticAlgorithm):

    def __init__(self,
                 pop_size=50,
                 n_offsprings=None,
                 output=SingleObjectiveOutput(),
                 sampling=MixedVariableSampling(),
                 mating=MixedVariableMating(eliminate_duplicates=MixedVariableDuplicateElimination()),
                 eliminate_duplicates=MixedVariableDuplicateElimination(),
                 survival=FitnessSurvival(),
                 **kwargs):
        super().__init__(pop_size=pop_size, n_offsprings=n_offsprings, sampling=sampling, mating=mating,
                         eliminate_duplicates=eliminate_duplicates, output=output, survival=survival, **kwargs)



"""

Differential Evolution (DE)

-------------------------------- Description -------------------------------



-------------------------------- References --------------------------------

[1] J. Blank and K. Deb, pymoo: Multi-Objective Optimization in Python, in IEEE Access,
vol. 8, pp. 89497-89509, 2020, DOI: 10.1109/ACCESS.2020.2990567

-------------------------------- License -----------------------------------


----------------------------------------------------------------------------
"""

import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.core.infill import InfillCriterion
from pymoo.core.population import Population
from pymoo.core.replacement import ImprovementReplacement
from pymoo.core.variable import Choice, get
from pymoo.core.variable import Real
from pymoo.docs import parse_doc_string
from pymoo.operators.control import EvolutionaryParameterControl, NoParameterControl
from pymoo.operators.crossover.binx import mut_binomial
from pymoo.operators.crossover.expx import mut_exp
from pymoo.operators.mutation.pm import PM
from pymoo.operators.repair.bounds_repair import repair_random_init
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.rnd import fast_fill_random
from pymoo.termination.default import DefaultSingleObjectiveTermination
from pymoo.util.display.single import SingleObjectiveOutput
from pymoo.util.misc import where_is_what


# =========================================================================================================
# Crossover
# =========================================================================================================

def de_differential(X, F, jitter, alpha=0.001):
    n_parents, n_matings, n_var = X.shape
    assert n_parents % 2 == 1, "For the differential an odd number of values need to be provided"

    # the differentials from each pair
    delta = np.zeros((n_matings, n_var))

    # for each difference of the differences
    for i in range(1, n_parents, 2):
        # create the weight vectors with jitter to give some variation
        _F = F[:, None].repeat(n_var, axis=1)
        _F[jitter] *= (1 + alpha * (np.random.random((jitter.sum(), n_var)) - 0.5))

        # add the difference to the vector
        delta += _F * (X[i] - X[i + 1])

    # now add the differentials to the first parent
    Xp = X[0] + delta

    return Xp


# =========================================================================================================
# Variant
# =========================================================================================================

class Variant(InfillCriterion):

    def __init__(self,
                 selection="best",
                 n_diffs=1,
                 F=0.5,
                 crossover="bin",
                 CR=0.2,
                 jitter=False,
                 prob_mut=0.1,
                 control=EvolutionaryParameterControl,
                 **kwargs):

        super().__init__(**kwargs)
        self.selection = Choice(selection, options=["rand", "best"], all=["rand", "best", "target-to-best"])
        self.n_diffs = Choice(n_diffs, options=[1], all=[1, 2])
        self.F = Real(F, bounds=(0.4, 0.7), strict=(0.0, None))
        self.crossover = Choice(crossover, ["bin"], all=["bin", "exp", "hypercube", "line"])
        self.CR = Real(CR, bounds=(0.2, 0.8), strict=(0.0, 1.0))
        self.jitter = Choice(jitter, options=[False], all=[True, False])

        self.mutation = PM(at_least_once=True)
        self.mutation.eta = 20
        self.mutation.prob = prob_mut

        self.control = control(self)

    def do(self, problem, pop, n_offsprings, algorithm=None, **kwargs):
        control = self.control

        # let the parameter control now some information
        control.tell(pop=pop)

        # set the controlled parameter for the desired number of offsprings
        control.do(n_offsprings)

        # find the different groups of selection schemes and order them by category
        sel, n_diffs = get(self.selection, self.n_diffs, size=n_offsprings)
        H = where_is_what(zip(sel, n_diffs))

        # get the parameters used for reproduction during the crossover
        F, CR, jitter = get(self.F, self.CR, self.jitter, size=n_offsprings)

        # the `target` vectors which will be recombined
        X = pop.get("X")

        # the `donor` vector which will be obtained through the differential equation
        donor = np.full((n_offsprings, problem.n_var), np.nan)

        # for each type defined by the type and number of differentials
        for (sel_type, n_diffs), targets in H.items():

            # the number of offsprings created in this run
            n_matings, n_parents = len(targets), 1 + 2 * n_diffs

            # create the parents array
            P = np.full([n_matings, n_parents], -1)

            itself = np.array(targets)[:, None]

            best = lambda: np.random.choice(np.where(pop.get("rank") == 0)[0], replace=True, size=n_matings)

            if sel_type == "rand":
                fast_fill_random(P, len(pop), columns=range(n_parents), Xp=itself)
            elif sel_type == "best":
                P[:, 0] = best()
                fast_fill_random(P, len(pop), columns=range(1, n_parents), Xp=itself)
            elif sel_type == "target-to-best":
                P[:, 0] = targets
                P[:, 1] = best()
                fast_fill_random(P, len(pop), columns=range(2, n_parents), Xp=itself)
            else:
                raise Exception("Unknown selection method.")

            # get the values of the parents in the design space
            XX = np.swapaxes(X[P], 0, 1)

            # do the differential crossover to create the donor vector
            Xp = de_differential(XX, F[targets], jitter[targets])

            # make sure everything stays in bounds
            if problem.has_bounds():
                Xp = repair_random_init(Xp, XX[0], *problem.bounds())

            # set the donors (the one we have created in this step)
            donor[targets] = Xp

        # the `trial` created by by recombining target and donor
        trial = np.full((n_offsprings, problem.n_var), np.nan)

        crossover = get(self.crossover, size=n_offsprings)
        for name, K in where_is_what(crossover).items():

            _target = X[K]
            _donor = donor[K]
            _CR = CR[K]

            if name == "bin":
                M = mut_binomial(len(K), problem.n_var, _CR, at_least_once=True)
                _trial = np.copy(_target)
                _trial[M] = _donor[M]
            elif name == "exp":
                M = mut_exp(n_offsprings, problem.n_var, _CR, at_least_once=True)
                _trial = np.copy(_target)
                _trial[M] = _donor[M]
            elif name == "line":
                w = np.random.random((len(K), 1)) * _CR[:, None]
                _trial = _target + w * (_donor - _target)
            elif name == "hypercube":
                w = np.random.random((len(K), _target.shape[1])) * _CR[:, None]
                _trial = _target + w * (_donor - _target)
            else:
                raise Exception(f"Unknown crossover variant: {name}")

            trial[K] = _trial

        # create the population
        off = Population.new(X=trial)

        # do the mutation which helps to add some more diversity
        off = self.mutation(problem, off)

        # repair the individuals if necessary - disabled if repair is NoRepair
        off = self.repair(problem, off, **kwargs)

        # advance the parameter control by attaching them to the offsprings
        control.advance(off)

        return off


# =========================================================================================================
# Implementation
# =========================================================================================================


class DE(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 n_offsprings=None,
                 sampling=FloatRandomSampling(),
                 variant="DE/best/1/bin",
                 output=SingleObjectiveOutput(),
                 **kwargs
                 ):

        if variant is None:
            if "control" not in kwargs:
                kwargs["control"] = NoParameterControl
            variant = Variant(**kwargs)

        elif isinstance(variant, str):
            try:
                _, selection, n_diffs, crossover = variant.split("/")
                if "control" not in kwargs:
                    kwargs["control"] = NoParameterControl
                variant = Variant(selection=selection, n_diffs=int(n_diffs), crossover=crossover, **kwargs)
            except:
                raise Exception("Please provide a valid variant: DE/<selection>/<n_diffs>/<crossover>")

        super().__init__(pop_size=pop_size,
                         n_offsprings=n_offsprings,
                         sampling=sampling,
                         mating=variant,
                         survival=None,
                         output=output,
                         eliminate_duplicates=False,
                         **kwargs)

        self.termination = DefaultSingleObjectiveTermination()

    def _initialize_advance(self, infills=None, **kwargs):
        FitnessSurvival().do(self.problem, self.pop, return_indices=True)

    def _infill(self):
        infills = self.mating.do(self.problem, self.pop, self.n_offsprings, algorithm=self)

        # tag each individual with an index - if a steady state version is executed
        index = np.arange(len(infills))

        # if number of offsprings is set lower than pop_size - randomly select
        if self.n_offsprings < self.pop_size:
            index = np.random.permutation(len(infills))[:self.n_offsprings]
            infills = infills[index]

        infills.set("index", index)

        return infills

    def _advance(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must to be provided."

        # get the indices where each offspring is originating from
        I = infills.get("index")

        # replace the individuals with the corresponding parents from the mating
        self.pop[I] = ImprovementReplacement().do(self.problem, self.pop[I], infills)

        # update the information regarding the current population
        FitnessSurvival().do(self.problem, self.pop, return_indices=True)

    def _set_optimum(self, **kwargs):
        k = self.pop.get("rank") == 0
        self.opt = self.pop[k]


parse_doc_string(DE.__init__)



# -*- encoding: utf-8 -*-

import inspect
import re
from pymoo.config import Config

# =========================================================================================================
# Docstrings Dictionary
# =========================================================================================================


algorithms = {
    "pop_size": """ int
                The population sized used by the algorithm.
    """,

    "n_offsprings": """ int
                The number of offsprings that should be created in each generation.
    """,

    "sampling": """
                :class:`~pymoo.core.sampling.Sampling`, :class:`~pymoo.core.population.Population`, :obj:`numpy.array`
                    The sampling process defines the initial set of solutions which are the starting point of the
                    optimization algorithm. Here, you have three different options by passing

                        (i) A :class:`~pymoo.core.sampling.Sampling` implementation which is an implementation of a 
                        random sampling method. 

                        (ii) A :class:`~pymoo.core.population.Population` object containing the variables to
                        be evaluated initially OR already evaluated solutions (F needs to be set in this case).

                        (iii) Pass a two dimensional :obj:`numpy.array` with (n_individuals, n_var) which contains the variable 
                        space values for each individual.        
                """,

    "selection": """:class:`~pymoo.core.selection.Selection`
                    This object defines the mating selection to be used. 
                    In an evolutionary algorithm each generation parents need to be selected to produce new offsprings using 
                    different recombination and mutation operators. Different strategies for selecting parents are possible e.g. 
                    selecting them just randomly, only in the neighborhood, using a tournament selection to introduce some selection 
                    pressure, ... 
                    """,
    "crossover": """:class:`~pymoo.core.crossover.Crossover`
                    The crossover has the purpose of create offsprings during the evolution. After the mating selection
                    the parents are passed to the crossover operator which will dependent on the implementation create
                    a different number of offsprings.
                """,

    "mutation": """:class:`~pymoo.core.mutation.Mutation`
                    Some genetic algorithms rely only on the mutation operation. However, it has shown that increases
                    the performance to perform a mutation after creating the offsprings through crossover as well.
                    Usually the mutation operator needs to be initialized with a probability to be executed. 
                    Having a high probability of mutation will most of the time increase the diversity in the population.
                """,

    "survival": """:class:`~pymoo.core.survival.Survival`
                    The survival selection is the key for many genetic algorithms. It is responsible to define the
                    goal of convergence by choosing the individual to survive or be truncated each generation.
                    For single-objective single a selection based on the fitness is used commonly. However, for 
                    multi-objective single different concepts are introduced.
                """,

    "ref_points": """:obj:`numpy.array`
                    Reference Points (or also called Aspiration Points) as a :obj:`numpy.array` where each row 
                    represents a point and each column a variable (must be equal to the objective dimension of the problem)
                """,

    "eliminate_duplicates": """bool
                    The genetic algorithm implementation has a built in feature that eliminates duplicates after merging
                    the parent and the offspring population. If there are duplicates with respect to the current 
                    population or in the offsprings itself they are removed and the mating process is repeated to
                    fill up the offsprings until the desired number of unique offsprings is met.            
                """,

    "n_offsprings": """int (default: None)
                    Number of offspring that are created through mating. By default *n_offsprings=None* which
                    sets the number of offsprings equal to the population size. By setting *n_offsprings=1* a, so called,
                    steady-state version of an algorithm can be achieved.
            """,

    "ref_dirs": """:obj:`numpy.array`
                    The reference direction that should be used during the optimization. Each row represents a reference line
                    and each column a variable.
            """

}

visualization = {
    "figsize": """tuple
                The figure size. Default (figsize=(8, 6)). For some plots changing the size might have side-effects for position.
                """,

    "title": """str or tuple
                The title of the figure. If some additional kwargs should be provided this can be achieved by providing a tuple
                ("name", {"key" : val}).
                """,

    "legend": """str
                Whether a legend should be shown or not.
                """,

    "tight_layout": """bool
                        Whether tight layout should be used.
                    """,

    "bounds": """tuple
                If plot requires normalization, it might be necessary to supply the boundaries. (Otherwise they might be
                approximate by the minimum and maximum of the provided data). The boundaries should be provided as a list/tuple or
                2D numpy array, where the first element represents the minimum, second the second the maximum values.
                If only an integer or float is supplied, the boundaries apply for each variable.
            """,

    "reverse": """bool
                    If plot requires normalization, then the reverse values can be plotted (1 - Input). For some plots
                    it can be useful to interpret a larger area as better regarding a value. If minimization applies, a smaller
                    area means better, which can be misleading.
                """,

    "axis_style": """dict
                        Most of the plots consists of an axis. The style of the axis, e.g. color, alpha, ..., can be changed to
                        further modify the plot appealing.
                    """,

    "cmap": """colormap
                    For some plots different kind of colors are used. The colormap can be changed to modify the color sequence
                    for the plots.
            """,

    "labels": """str or list
                    The labels to be used for each variable provided in the plot. If a string is used, then they will
                    be enumerated. Otherwise, a list equal to the number of variables can be provided directly.
            """,

    "func_number_to_text": """func
                                A function which defines how numerical values should be represented if present in the plot 
                                for instance scientific notation, rounding and so on.
                            """,

}

docs = {**algorithms, **visualization}


# =========================================================================================================
# Util for docstrings
# =========================================================================================================

def parse_doc_string(source, dest=None, other={}):
    if not Config.parse_custom_docs:
        return

    if dest is None:
        dest = source

    D = {k: v.strip() for k, v in docs.items()}
    
    doc = source.__doc__
    

    lines = inspect.getsource(source)
    if doc is not None:
        
        doc = doc.format(**{**D, **other})

        lines = inspect.getsource(source)

        cnt = 0
        b = False

        for i, c in enumerate(lines):
            
            if b and cnt == 0:
                break
            if c == "(":
                cnt += 1
                b = True
            elif c == ")":
                cnt -= 1

        signature = lines[:i]
        signature = re.sub(r"[\n\t]*", "", signature)
        signature = re.sub(r"\s+", " ", signature)
        signature = re.sub(r"def\s*", "", signature)
        signature = signature.strip()

        if dest is not None:
            dest.__doc__ = signature + "\n" + doc



"""

Particle Swarm Optimization (PSO)

-------------------------------- Description -------------------------------



-------------------------------- References --------------------------------

[1] J. Blank and K. Deb, pymoo: Multi-Objective Optimization in Python, in IEEE Access,
vol. 8, pp. 89497-89509, 2020, DOI: 10.1109/ACCESS.2020.2990567

-------------------------------- License -----------------------------------


----------------------------------------------------------------------------
"""

import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.core.duplicate import NoDuplicateElimination
from pymoo.core.infill import InfillCriterion
from pymoo.core.population import Population
from pymoo.core.replacement import ImprovementReplacement, is_better
from pymoo.core.variable import Real, Choice, get
from pymoo.docs import parse_doc_string
from pymoo.operators.control import EvolutionaryParameterControl
from pymoo.operators.repair.bounds_repair import repair_random_init, repair_clamp
from pymoo.operators.sampling.rnd import random_by_bounds, FloatRandomSampling
from pymoo.termination.default import DefaultSingleObjectiveTermination
from pymoo.util.display.single import SingleObjectiveOutput
from pymoo.util.sliding_window import SlidingWindow


# =========================================================================================================
# Mating
# =========================================================================================================


def pso_canonical(V, X, P_X, L_X, w, c1, c2):
    n_particles, n_var = X.shape
    r1, r2 = np.random.random((n_particles, n_var)), np.random.random((n_particles, n_var))
    Vp = w * V + c1 * r1 * (P_X - X) + c2 * r2 * (L_X - X)
    return Vp


def pso_rotation_invariant(V, X, P_X, L_X, inertia, c1, c2):
    n_particles, n_var = X.shape

    r1 = np.random.random((n_particles, n_var))
    p = X + c1 * r1 * (P_X - X)

    r2 = np.random.random((n_particles, n_var))
    l = X + c2 * r2 * (L_X - X)

    G = (X + p + l) / 3
    r = np.linalg.norm(G - X, axis=1, keepdims=True)

    Vp = inertia * V + alea_sphere(G, r) - X

    return Vp


def alea_sphere(G, radius):
    n, m = G.shape

    x = np.random.normal(size=(n, m))
    l = np.sqrt(np.sum(x ** 2, axis=1, keepdims=True))

    r = np.random.random(size=(n, 1))
    x = r * radius * x / l
    return x + G


class Swarm(InfillCriterion):

    def __init__(self,
                 w=0.7,
                 c1=1.4,
                 c2=1.4,
                 V_max=0.2,
                 prob_mut=0.33,
                 control=EvolutionaryParameterControl,
                 **kwargs):

        super().__init__(**kwargs)
        self.w = Real(w, bounds=(0.7, 0.9), strict=(0.0, 1.0))
        self.c1 = Real(c1, bounds=(1.2, 1.6), strict=(0.0, None))
        self.c2 = Real(c2, bounds=(1.2, 1.6), strict=(0.0, None))
        self.V_max = V_max
        self.prob_mut = prob_mut

        # of parameter control should be applied on the mating level
        self.control = control(self)

    def do(self, problem, pop, n_offsprings, algorithm=None, **kwargs):
        control = self.control

        # let the parameter control now some information
        control.tell(pop=pop)

        # set the controlled parameter for the desired number of offsprings
        control.do(n_offsprings)

        # get the parameters that will be used
        w, c1, c2 = get(self.w, self.c1, self.c2, size=(len(pop), 1))

        # get all the population that play a role for the mating
        swarm, pbest, lbest = algorithm.swarm, algorithm.pbest, algorithm.lbest

        V, X, P_X, L_X = swarm.get("V"), swarm.get("X"), pbest.get("X"), lbest.get("X")

        Vp = pso_canonical(V, X, P_X, L_X, w, c1, c2)
        # Vp = pso_rotation_invariant(V, X, P_X, L_X, w, c1, c2)

        # if a maximum velocity has been defined
        V_max = self.V_max
        if V_max is not None:
            xl, xu = problem.bounds()
            Vp = repair_clamp(Vp, -V_max * (xu - xl), V_max * (xu - xl))

        # the position of the new swarm particles
        Xp = X + Vp

        # if adding the velocity has brought them out of bounds -> bring them back
        if problem.has_bounds():
            Xp = repair_random_init(Xp, X, *problem.bounds())

        # do a mutation  of the global best solution (helps to keep some diversity)
        # Xm = PM(prob=1.0, eta=20).do(problem, swarm).get("X")
        # mut = pbest.get("rank") == 0
        # Xp[mut] = Xm[mut]

        # recalculate the velocity after the repair has happened
        Vp = Xp - X

        # create the population
        off = Population.new(X=Xp, V=Vp)

        # do the reset of particles if their personal bests have not moved much
        # for k, ind in enumerate(pbest):
        #     delta = algorithm.delta[k]
        #
        #     if k != algorithm.best and delta.is_full() and np.array(delta).mean() < 0.001:
        #         particle = FloatRandomSampling().do(problem, 1)[0]
        #         particle.set("V", np.zeros(problem.n_var))
        #         off[k], pbest[k], lbest[k] = particle, particle, particle
        #         delta.clear()

        # repair the individuals if necessary - disabled if repair is NoRepair
        off = self.repair(problem, off, **kwargs)

        # advance the parameter control by attaching them to the offsprings
        control.advance(off)

        return off


def get_neighbors(name, N):
    if name == "star":
        return np.tile(np.arange(N), (N, 1))
    elif name == "ring":
        return (np.array([np.arange(3) for _ in range(N)]) + np.arange(N)[:, None] - 1) % N
    elif name.startswith("random"):
        K = 3
        neighbors = []
        for i in range(N):
            vals = np.random.permutation(N)[:K]
            neighbors.append([i] + vals.tolist())
        return neighbors
    else:
        raise Exception(f"Unknown topology: {name}")


# =========================================================================================================
# Implementation
# =========================================================================================================


class EPPSO(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=FloatRandomSampling(),
                 swarm=Swarm(),
                 topology="star",
                 init_V="zero",
                 output=SingleObjectiveOutput(),
                 **kwargs):

        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         mating=swarm,
                         init_V=init_V,
                         n_offsprings=None,
                         eliminate_duplicates=NoDuplicateElimination(),
                         output=output,
                         **kwargs)

        # how the initial weights should be created
        self.init_V = Choice(init_V, options=["zero", "random"])

        # how the individual are connected to determine the local (or also global) best
        self.topology = Choice(topology, options=["star", "ring"])

        # create the neighbors of each particle given the topology
        self.neighbors = get_neighbors(get(self.topology), pop_size)

        # choose the single-objective default termination
        self.termination = DefaultSingleObjectiveTermination()

        # the particles that fly around to find good solutions (pop is the pbest)
        self.swarm = None

        # the personal and local best solution
        self.lbest = None
        self.pbest = None

        # the integer of the currently best individual
        self.best = None

    def _initialize_infill(self):
        swarm = super()._initialize_infill()

        n_var = self.problem.n_var
        init_V = get(self.init_V)
        if init_V == "zero":
            V = np.zeros((len(swarm), n_var))
        elif init_V == "random":
            xl, xu = self.problem.bounds()
            Xp = random_by_bounds(n_var, xl, xu, len(swarm))
            V = (swarm.get("X") - Xp) / 2
        else:
            raise Exception("Unknown velocity initialization.")
        swarm.set("V", V)

        return swarm

    def _initialize_advance(self, infills=None, **kwargs):
        self.swarm = infills
        self.pbest = self.pop
        self.lbest = Population.create(*self.pbest)
        self.delta = [SlidingWindow(30) for _ in range(len(infills))]

        FitnessSurvival().do(self.problem, self.pbest, return_indices=True)

    def _advance(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus 'infills' must to be provided."

        X = self.pbest.get("X")

        self.swarm = infills
        ImprovementReplacement().do(self.problem, self.pbest, infills, inplace=True)
        Xp = self.pbest.get("X")

        xl, xu = self.problem.bounds()
        delta = np.max(np.abs(X - Xp) / (xu - xl), axis=1)
        [self.delta[k].append(delta[k]) for k in range(len(delta))]

        pbest = self.pbest
        S = FitnessSurvival().do(self.problem, pbest, return_indices=True)
        rank = pbest.get("rank")
        self.best = S[0]

        if get(self.topology) == "random-adaptive" and pbest[self.best].get("n_gen") != self.n_gen:
            self.neighbors = get_neighbors(get(self.topology), len(pbest))

        # send the message from each particle to all its neighbors
        msgs = [[] for _ in range(len(pbest))]
        for k, neighbors in enumerate(self.neighbors):
            for neighbor in neighbors:
                msgs[neighbor].append(k)

        # now receive the messages and set the new local best (if an improvement has been found)
        for k, msg in enumerate(msgs):

            # if messages have been received
            if len(msg) > 0:

                # find the best one from the swarm that have been send
                i = msg[rank[msg].argmin()]

                # if the best from the message is better than the current local best
                if is_better(pbest[i], self.lbest[k]):
                    self.lbest[k] = pbest[i]

        self.pop = self.pbest

    def _set_optimum(self, **kwargs):
        k = self.pop.get("rank") == 0
        self.opt = self.pop[k]


parse_doc_string(EPPSO.__init__)




import os
import subprocess

import numpy as np

from pymoo.core.problem import ElementwiseProblem


class MOPTA08(ElementwiseProblem):
    def __init__(self, exec):
        super().__init__(n_var=124, n_obj=1, n_ieq_constr=68, xl=0, xu=1, vtype=float)
        self.exec = exec

    def _evaluate(self, x, out, *args, **kwargs):
        inputs = os.linesep.join([str(e) for e in x])
        res = subprocess.run(self.exec, input=inputs, text=True, stdout=subprocess.PIPE)
        val = np.array([float(e) for e in res.stdout.split()])
        out["F"] = val[0]
        out["G"] = val[1:]

    def _calc_pareto_front(self, *args, **kwargs):
        return 222.74







import json
import os
import urllib.request
from os.path import join, dirname, abspath

import numpy as np

from pymoo.config import Config


class Remote:
    # -------------------------------------------------
    # Singleton Pattern
    # -------------------------------------------------
    __instance = None

    @staticmethod
    def get_instance():
        if Remote.__instance is None:
            server = Config.data()
            folder = join(dirname(dirname(abspath(__file__))), 'data')
            Remote.__instance = Remote(server, folder)
        return Remote.__instance

    # -------------------------------------------------

    def __init__(self, server, folder=None) -> None:
        super().__init__()
        self.server = server
        self.folder = folder

    def load(self, *args, to="numpy"):

        # the local file we can try loading
        f = join(str(self.folder), *args)

        # check if that path already exists
        if not os.path.exists(f):

            # if not make sure to create it that the file can be written
            folder = dirname(f)
            if not os.path.exists(folder):
                os.makedirs(folder)

            # create the url to load it from and download the file remotely
            url = self.server + "/".join(args)
            urllib.request.urlretrieve(url, f)

        if to == "numpy":
            return np.loadtxt(f)
        elif to == "json":
            with open(f) as file:
                return json.load(file)

        return f



import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem


class BNH(Problem):

    def __init__(self):
        super().__init__(n_var=2, n_obj=2, n_ieq_constr=2, vtype=float)
        self.xl = np.zeros(self.n_var)
        self.xu = np.array([5.0, 3.0])

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = 4 * x[:, 0] ** 2 + 4 * x[:, 1] ** 2
        f2 = (x[:, 0] - 5) ** 2 + (x[:, 1] - 5) ** 2
        g1 = (1 / 25) * ((x[:, 0] - 5) ** 2 + x[:, 1] ** 2 - 25)
        g2 = -1 / 7.7 * ((x[:, 0] - 8) ** 2 + (x[:, 1] + 3) ** 2 - 7.7)

        out["F"] = anp.column_stack([f1, f2])
        out["G"] = anp.column_stack([g1, g2])

    def _calc_pareto_front(self, n_points=100):
        x1 = np.linspace(0, 5, n_points)
        x2 = np.linspace(0, 5, n_points)
        x2[x1 >= 3] = 3

        X = np.column_stack([x1, x2])
        return self.evaluate(X, return_values_of=["F"])








import pymoo.gradient.toolbox as anp
import numpy as np

from pymoo.core.problem import Problem


class Rastrigin(Problem):
    def __init__(self, n_var=2, A=10.0):
        super().__init__(n_var=n_var, n_obj=1, xl=-5, xu=5, vtype=float)
        self.A = A

    def _evaluate(self, x, out, *args, **kwargs):
        z = anp.power(x, 2) - self.A * anp.cos(2 * anp.pi * x)
        out["F"] = self.A * self.n_var + anp.sum(z, axis=1)

    def _calc_pareto_front(self):
        return 0.0

    def _calc_pareto_set(self):
        return np.full(self.n_var, 0)



from math import log

from pymoo.core.termination import Termination


class IndicatorTermination(Termination):

    def __init__(self, indicator, threshold, goal, **kwargs) -> None:
        super().__init__()

        # the indicator to be used
        self.indicator = indicator

        # define the threshold for termination
        self.threshold = threshold

        # what is the optimization goal for this indicator
        self.goal = goal
        assert goal in ["minimize", "maximize"]

        # optional parameters when the indicator calculation is performed
        self.kwargs = kwargs

        # initial the minimum and maximum values of the indicator
        self._min = float("inf")
        self._max = -float("inf")

    def _update(self, algorithm):

        # get the objective space values
        F = algorithm.opt.get("F")

        # get the resulting value from the indicator
        v = self.indicator.do(F, **self.kwargs)

        threshold = self.threshold

        # update the minimum and maximum boundary ranges
        self._min = min(self._min, v)
        self._max = max(self._max, v)
        _min, _max = self._min, self._max

        # depending on the goal either set the percentage
        if self.goal == "minimize":
            perc = 1 - (v - threshold) / (_max - threshold)
        else:
            perc = (v - _min) / (threshold - _min)

        return perc



from pymoo.core.decomposition import Decomposition
from pymoo.util.function_loader import load_function


class PBI(Decomposition):

    def __init__(self, theta=5, **kwargs) -> None:
        super().__init__(**kwargs)
        self.theta = theta

    def _do(self, F, weights, **kwargs):
        d1, d2 = load_function("calc_distance_to_weights")(F, weights, self.utopian_point)
        return d1 + self.theta * d2



from abc import abstractmethod, ABC

from pymoo.core.callback import Callback
from pymoo.core.problem import Problem


class DynamicProblem(Problem, ABC):
    pass


class DynamicTestProblem(DynamicProblem, ABC):

    def __init__(self, nt, taut, tau=1, time=None, **kwargs):
        super().__init__(**kwargs)
        self.tau = tau
        self.nt = nt
        self.taut = taut
        self._time = time

    def tic(self, elapsed=1):

        # increase the time counter by one
        self.tau += elapsed

        # remove the cache of the problem to recreate ps and pf
        self.__dict__["cache"] = {}

    @property
    def time(self):
        if self._time is not None:
            return self._time
        else:
            return 1 / self.nt * (self.tau // self.taut)

    @time.setter
    def time(self, value):
        self._time = value


class TimeSimulation(Callback):

    def update(self, algorithm):
        problem = algorithm.problem
        if hasattr(problem, "tic"):
            problem.tic()
        else:
            raise Exception("TimeSimulation can only be used for dynamic test problems.")



import numpy as np

from pymoo.core.crossover import Crossover
from pymoo.core.variable import Real, get
from pymoo.util.misc import row_at_least_once_true


def mut_binomial(n, m, prob, at_least_once=True):
    prob = np.ones(n) * prob
    M = np.random.random((n, m)) < prob[:, None]

    if at_least_once:
        M = row_at_least_once_true(M)

    return M


class BinomialCrossover(Crossover):

    def __init__(self, bias=0.5, n_offsprings=2, **kwargs):
        super().__init__(2, n_offsprings, **kwargs)
        self.bias = Real(bias, bounds=(0.1, 0.9), strict=(0.0, 1.0))

    def _do(self, problem, X, **kwargs):
        _, n_matings, n_var = X.shape

        bias = get(self.bias, size=n_matings)
        M = mut_binomial(n_matings, n_var, bias, at_least_once=True)

        if self.n_offsprings == 1:
            Xp = X[0].copy(X)
            Xp[~M] = X[1][~M]
        elif self.n_offsprings == 2:
            Xp = np.copy(X)
            Xp[0][~M] = X[1][~M]
            Xp[1][~M] = X[0][~M]
        else:
            raise Exception

        return Xp


class BX(BinomialCrossover):
    pass



import numpy as np

from pymoo.core.meta import Meta
from pymoo.core.problem import Problem


def calc_complex_gradient(problem, return_values_of, x, eps, *args, **kwargs):
    xp = x + np.eye(len(x)) * np.complex(0, eps)
    out = problem.do(xp, return_values_of, *args, **kwargs)

    grad = {}
    for name, value in out.items():
        grad[name] = np.imag(value / eps).T

    return grad


class ComplexNumberGradient(Meta, Problem):

    def __init__(self, object, eps=1e-8, **kwargs):
        super().__init__(object, **kwargs)
        self.eps = eps

    def do(self, X, return_values_of, *args, **kwargs):
        out = self.__object__.do(X, return_values_of, *args, **kwargs)

        vals_not_grad = [v for v in return_values_of if not v.startswith("d")]

        for i, x in enumerate(X):
            grad = calc_complex_gradient(self.__object__, vals_not_grad, x, self.eps, *args, **kwargs)

            for name, value in grad.items():
                out['d' + name][i] = value

        return out



import numpy as np

from pymoo.core.algorithm import Algorithm
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.core.replacement import is_better
from pymoo.termination.default import DefaultSingleObjectiveTermination


class ExponentialSearch(Algorithm):

    def __init__(self, delta=0.05, **kwargs):
        super().__init__(**kwargs)
        self.termination = DefaultSingleObjectiveTermination()
        self.alpha = delta
        self.point = None

    def _setup(self, problem, x0=None, **kwargs):
        msg = "Only problems with one variable, one objective and no constraints can be solved!"
        assert problem.n_var == 1 and not problem.has_constraints() and problem.n_obj == 1, msg
        self.point = x0

    def _initialize_infill(self):
        self.step_size = self.alpha

        if self.point is None:
            return Population.new(X=np.copy(self.problem.xl[None, :]))
        else:
            return Population.create(self.point)

    def step(self):
        alpha, max_alpha = self.alpha, self.problem.xu[0]

        if alpha > max_alpha:
            alpha = max_alpha

        infill = Individual(X=np.array([alpha]))
        self.evaluator.eval(self.problem, infill)
        self.pop = Population.merge(self.pop, infill)[-10:]

        if is_better(self.point, infill, eps=0.0) or alpha == max_alpha:
            self.termination.force_termination = True
            return

        self.point = infill
        self.alpha *= 2
